	STITLE	ALG MODULAR ARITHMETIC
**********************************************************************
*	ALG Library:	Modular arithmetic
**********************************************************************

**********************************************************************
*
* If Z is a coefficient in modular (mod Zn) mathematics then
*
*   in positive modular math 0 <= Z < Zn
*   in symmetric modular math -Zn/2 <= Z <= Zn/2
*
* Positive modular math is commonly used with large Zn in integer
* factorization, Zn being the integer to be factored.
*
* Symmetric modular math is used in polynomial mathematics with
* small Zn. Here the small modulo is used to speed up intermediate
* calculations in Z/Zn until the results are then 'lifted' to a
* higher modulo to get actual results in Z.
*
* Positive modular math subroutines are indicated by having
* a leading 'Mod' in the subroutine names while in symmetric
* modular math 'Mod' is in the end of the names.
*
* Correspondingly subroutines exists for the two types only as needed.
*
**********************************************************************

**********************************************************************
* EXPORT:
*
* Positive modular math:
*
*	Mod		(    Zb Zn --> Z  )
*	ModAdd		( Za Zb Zn --> Z  )
*	ModSub		( Za Zb Zn --> Z  )
*	ModMul		( Za Zb Zn --> Z  )
*	ModDiv		( Za Zb Zn --> Z  )
*	ModInv		(    Zb Zn --> Z  )
*	ModPow		( Zb Zp Zn --> Z  )
*
* Symmetric modular math:
*
*	QMod		(     Q Zn --> Q' )
*	QAddMod		( Q1 Q2 Zn --> Q  )
*	QSubMod		( Q1 Q2 Zn --> Q  )
*	QMulMod		( Q1 Q2 Zn --> Q  )
*	QDivMod		( Q2 Q2 Zn --> Qquo Qrem )
*	QInvMod		(     Q Zn --> Q' )
*	QGcdMod		( Q1 Q2 Zn --> Q  )
*	QGcdExMod	( Q1 Q2 Zn --> Q1' Q2' )
*
*	PEvalMod	(   Q Z Zn --> Q' )
*
**********************************************************************

**********************************************************************
*		POSITIVE MODULAR MATH
**********************************************************************

ASSEMBLE
sSHORTMOD	EQU 9		Help flag for modular math
RPL


**********************************************************************
* Desc:		Modular power.		Z = Zb^Zp (mod Zn)
* Stack:	( Zb Zp Zn --> Z' )
* Notes:	Zp power, not system binary!
*
*	This is the main bottle-neck of most algorithms involving
*	primes. For example Pollard's p-1 factorization with the
*	first 1000 primes would require calculating
*
*		2^lcm(1..7919) mod N
*
*	In practise this is done in steps of maybe 50, but still
*	gives a clear indication of the time required in the bad cases.
*		The LCM takes 1414.5 bytes as a Z
*		Calculating gcd(n,2^z-1) would provide a very simple
*		factorization program..
*
*	Another example is the BRabin subroutine, which simply had to
*	be limited to a smaller base-set to reduce execution time for
*	very long integers.
*
**********************************************************************
NULLNAME ModPow
::
* Special case ^0

  OVER QIsZero? casedrop
	:: 2DROP HXS 2 10 ;		( --> Z1 )

* Handle negative power

  OVER ZIsNeg? IT
  :: SWAP ZAbs SWAP
     ROTOVER ModInv UNROT
  ;

* Normalize to 0 <= Zb < Zn

  ROTOVER				( Zp Zn Zb Zn )
  Mod					( Zp Zn Zb' )

* Special case ^1

  UNROTSWAP				( Zb Zn Zp )
  DupZIsOne? case2DROP			( --> Zb )

* Logarithmic multiplication loop in assembly follows
* The code expects all arguments to be normalized.

* ( Zb Zn Zp --> Zb^Zp mod Zn )
  CODE
		GOSBVL	=SAVPTR

		GOSUB	D1->Zn
		D1=D1+	5
		A=DAT1	A
		A=A+CON	A,5		Zn size
		R1=A	A		= slot size

		A=A+A	A		Slots:	1 for result R
		A=A+A	A			1 for multiplier P
		A=A+A	A			2 for *   2 for /	= 6
		LC(5)	400		Chicken
		C=C+A	A
		GOSBVL	=MAKE$N

* Count number of bits in the power

		A=DAT1	A		->Zp
		D0=A		
		GOSUBL	BitCount	A[A] = #bits
		C=A	A

* Initialize variable pointers

*	00	#bit	(loop index)
*	+5	#bits	(loop stop value)
*	+5	->R	(result)
*	+5	->P	(multiplier)
*	+5	->MUL	(temporary product)
*	+5	->DIV	(temporary division)

		GOSUB	D0->BIT
		A=0	A		Start from zero
		DAT0=A	A
		D0=D0+	5
		DAT0=C	A		End at #bits
		D0=D0+	5
		A=R1	A
		B=A	A		slot size
		A=R0	A		->$work
		A=A+CON	A,10
		DAT0=A	A		->R
		AD0EX
		LCZ1			Init R = 1
		DAT0=C	12
		AD0EX
		D0=D0+	5
		A=A+B	A
		DAT0=A	A		->P
		D0=D0+	5
		A=A+B	A
		DAT0=A	A		->MUL
		D0=D0+	5
		A=A+B	A
		A=A+B	A
		DAT0=A	A		->DIV

* Init P to be Zb

		GOSBVL	=D1=DSKTOP
		D1=D1+	10
		C=DAT1	A		->Zb
		GOSUB	D0->P
		D0=C			->Zb
		D1=A			->P
		GOSUB	CopyZ

* The main loop

MODPOWLOOP	GOSUB	D0->BIT
		A=DAT0	A		#bit
		GOSBVL	=D1=DSKTOP
		C=DAT1	A		->Zp
		GOSUBL	BitTest		CS if bit is set
		GONC	MODPOWNEXT

		GOSUB	D0->MUL		R = R*P mod N
		R3=A	A		->output
		GOSUB	D0->P
		D1=A
		GOSUB	D0->R
		GOSUBL	MULZZentry
		GOSUB	MODZn

		GOSUB	D0->R		Target = ->R
		D1=A
		GOSUB	CopyRem		Copy remainder to ->R

MODPOWNEXT	GOSUB	D0->BIT
		A=DAT0	A		#bit
		A=A+1	A		#bit++
		DAT0=A	A
		D0=D0+	5
		C=DAT0	A		#bits
		?A=C	A
		GOYES	MODPOWFINISH

		GOSUB	D0->MUL		P = P^2 mod N
		R3=A			->output
		GOSUB	D0->P
		D1=A
		GOSUBL	MULZZentry
		GOSUB	MODZn

		GOSUB	D0->P		Target = ->P
		D1=A
		GOSUB	CopyRem		Copy remainder to ->P

		GOTO	MODPOWLOOP

MODPOWFINISH	GOSUB	D0->R		->result
		D0=A
		A=A-CON	A,10
		R0=A			->$work
		GOSBVL	=SKIPOB		->result tail
		GOSBVL	=Shrink$	remove extra work memory
		GOSBVL	=GETPTR
		D1=D1+	10		Drop Zn and Zp
		D=D+1	A
		D=D+1	A

		A=R0	A		->$work
		A=A+CON	A,10		->result
		GOVLNG	=OverWrALoop	Overwrite ->Zb with ->Zr
		
*---------------------------------------------------------------------
MODZn		GOSUB	D0->DIV
		R0=A	A		->output
		GOSUB	D0->MUL
		GOSUB	D1->Zn
		GOLONG	ZQUO
*---------------------------------------------------------------------
CopyRem		GOSUB	D0->DIV		->quotient
		D0=D0+	5
		C=DAT0	A		quotient size
		AD0EX
		A=A+C	A		->quotient tail
		AD0EX
		D0=D0+	6		->remainder
CopyZ		D0=D0+	5
		C=DAT0	A
		D0=D0-	5
		C=C+CON	A,5
		GOVLNG	=MOVEDOWN
*---------------------------------------------------------------------
D1->Zn		GOSBVL	=D1=DSKTOP
		D1=D1+	5		->stack level 2
		C=DAT1	A
		D1=C			->Zn
		RTN
*---------------------------------------------------------------------
D0->DIV		GOSUB	D0->BIT		+25 from #bit
		D0=D0+	5
		GONC	+
*---------------------------------------------------------------------
D0->MUL		GOSUB	D0->BIT		+20 from #bit
+		D0=D0+	10
		GONC	+
*---------------------------------------------------------------------
D0->R		GOSUB	D0->BIT		+10 from #bit
+		D0=D0+	10
		GONC	+
*---------------------------------------------------------------------
D0->P		GOSUB	D0->BIT		+15 from #bit
		D0=D0+	15
+		A=DAT0	A
		D0=A
		RTN
*---------------------------------------------------------------------
D0->BIT		D0=(5)	(=IRAM@)-4
		A=DAT0	A
		D0=A
		D0=(4)	#100		#70100 or #80100 (in IRAMBUFF)
		RTN
  ENDCODE

;

*	* Logarithmic loop
*	
*	  ROTSWAP				( Zn Zb Zp )
*	  HXS 2 10 SWAP				( Zn Zb R Zp )  
*	
*	  ZBits					( Zn Zb R Zp #bits )
*	
*	  ZERO_DO (DO)				( Zn Zb R Zp )
*		INDEX@ ZBit?			( Zn Zb R Zp set? )
*		IT
*		::				( Zn Zb R Zp )
*		  SWAP3PICK QMul		( Zn Zb Zp R' )
*		  4PICK QDivRem SWAPDROPSWAP	( Zn Zb R'' Zp )
*		;
*		ISTOP-INDEX #1<> IT		( Square Zb if more to go )
*		::
*		  ROTDUP QMul			( Zn R Zp Zb' )
*		  4PICK QDivRem SWAPDROP UNROT	( Zn Zb' R Zp )
*		;
*	  LOOP
*	
*	  DROP UNROT2DROP
*	;
  

**********************************************************************
* Desc:		Make Z modulo N
* Stack:	( Z Zn --> Z' )
**********************************************************************
NULLNAME Mod
CODE
		GOSUBL	ModPrepare	D0 = ->Z	R4[W] = Z / ->Zn
		?ST=0	sSHORTMOD
		GOYES	ModCheckRPL

		C=R4	W		C[W] = Z
		GOSUBL	ZMODPSHORT	A[W] = Zmod	C[W] = Z
		GOLONG	PushN2Loop

ModCheckRPL	GOSUB	ModRPL:
ENDCODE
::
  2DUP ZNLT? ?SKIP				( Divide if too big )
  :: DUPUNROT QDivRem SWAPDROPSWAP ;
  OVER ZIsNeg? case QAdd			( Add N if negative )
  DROP						( Else just drop N )
;

**********************************************************************
* Desc:		Modular addition.	Z = Z1+Z2 (mod Zn)
* Stack:	( Z1 Z2 Zn --> Z )
**********************************************************************
NULLNAME ModAdd
CODE
		GOSUBL	ModPrepare	D0 = ->Z2	R4[W] = Zn / ->Zn
		?ST=0	sSHORTMOD
		GOYES	ModAddRPL
		C=R4	W		C[W] = Zn
		GOSUBL	ZMODPSHORT	A[W] = Z2mod	C[W] = Zn
modaddcont	R3=A	W
		D1=D1+	5
		A=DAT1	A
		D0=A
		GOSUBL	ZMODPSHORT	A[W] = Z1mod	C[W] = Zn
		C=R3	W
		A=A+C	W
		C=R4	W		C[W] = Zn
		GOC	+		Overflow - must fix then
		?A<C	W
		GOYES	++		Is fine
+		A=A-C	W
++		GOLONG	PushN3Loop

ModAddRPL	GOSUB	ModRPL:
ENDCODE
::
  UNROT QAdd SWAP		( Z' Zn )
  Mod
;
**********************************************************************
* Desc:		Modular substraction.	Z = Z1-Z2 (mod Zn)
* Stack:	( Z1 Z2 Zn --> Z )
**********************************************************************
NULLNAME ModSub
CODE
		GOSUBL	ModPrepare	D0 = ->Z2	R4[W] = Zn / ->Zn
		?ST=0	sSHORTMOD
		GOYES	ModSubRPL
		C=R4	W		C[W] = Zn
		GOSUBL	ZMODPSHORT
		R3=A	W
		A=A-C	W
		A=-A	W		"Negate"
		GOTO	modaddcont
ModRPL:		C=RSTK			Help subroutine to pack code
		GOVLNG	=GETPTREVALC
ModSubRPL	GOSUB	ModRPL:
ENDCODE
::
  UNROT QSub SWAP		( Z' Zn )
  Mod
;
**********************************************************************
* Desc:		Modular multiplication. Z = Z1*Z2 (mod Zn)
* Stack:	( Z1 Z2 Zn --> Z )	
**********************************************************************
NULLNAME ModMul
CODE
		GOSUBL	ModPrepare	D0 = ->Z2	R4[W] = Zn / ->Zn
		?ST=0	sSHORTMOD
		GOYES	ModMulRPL
		C=R4	W		C[W] = Zn
		GOSUBL	ZMODPSHORT
		R3=A	W		R3[W] = Z2mod
		D1=D1+	5
		A=DAT1	A
		D0=A
		GOSUBL	ZMODPSHORT	A[W] = Z1mod
		B=C	W		B[W] = Zn
		C=R3	W		C[W] = Z2mod
		GOSUBL	SHORTMULMOD	A[W] = Z1mod * Z2mod MOD Zn
		GOLONG	PushN3Loop

ModMulRPL	GOSUB	ModRPL:
ENDCODE
::
  UNROT QMul SWAP		( Z' Zn )
  Mod
;
**********************************************************************
* Desc:		Modular division. Z = Z1/Z2 (mod Zn)
* Stack:	( Z1 Z2 Zn --> Z )
**********************************************************************
NULLNAME ModDiv
::
  DUPUNROT		( Z1 Zn Z2 Zn )
  ModInv		( Z1 Zn Z' )
  SWAP ModMul		( Z'' )
;
**********************************************************************
* Desc:		Modular inversion. Z' = INV(Z) (mod Zn)
* Stack:	( Z Zn --> Z' )
* Notes:	Returns 0 if GCD[Z,Zn] <> 1 or if Z = 0
*		(Otherwise the results would be unpredictable.)
* Algorithm:	Z = Z MOD N
*		IF Z = 0 THEN RETURN 0
*		IF GCD(Z,N)<>1 THEN RETURN 0
*		R=1; S=0; P=N
*		WHILE P<>0
*		  Q=Z/P
*		  T=R-SQ
*		  R=S; S=T
*		  T=Z-PQ
*		  Z=P; P=T
*		REPEAT
*		IF R<0 THEN R=R+N
*		RETURN R
**********************************************************************
NULLNAME ModInv

CODE

* This code works only if above algorithm works mod N too.
* Step 1 is the only questionable one, Q=Z/P

		GOSUBL	ModPrepare	D0 = ->Z	R4[W] = Zn / ->Zn
		?ST=0	sSHORTMOD
		GOYES	ModInvRPL
		C=R4	W		C[W] = Zn
		?C=0	W
		GOYES	modinv0
		GOSUBL	ZMODINV		A[W] = Z inv Zn
		GOLONG	PushN2Loop
modinv0		A=0	W
		GOLONG	PushN2Loop	Return 0
ModInvRPL	GOSUB	ModRPL:
ENDCODE
::
  DupQIsZero? caseDROP			( --> Z )
  SWAP2DUP ZGcd				( Zn Z Zgcd )
  ZIsOne? NOTcase2drop HXS 1 0		( --> Z0 )
  SWAP					( x n )
  HXS 2 10				( x n r=1 )
  HXS 1 0				( x n r=1 s=0 )
  NULLHXS NULLHXS			( x n r=1 s=0 q t )
  5PICK					( x n r=1 s=0 q t p=n )
  {{ x n r s q t p }}

  ( p<>0 quaranteed on entry )
  BEGIN
	x p QDiv q!			( q=x/p )
	r s q QMul QSub t!		( t=r-sq )
	s r!				( r=s )
	t s!				( s=t )
	x p q QMul QSub t!		( t=x-pq )
	p x!				( x=p )
	t p!				( p=t )
	p QIsZero?			( p<>0 )
  UNTIL
  r DupZIsNeg?				( r<0 )
  IT :: n QAdd ;			( r=r+n )
  ABND
;

**********************************************************************
*		SYMMETRIC MODULAR MATH
**********************************************************************

**********************************************************************
* Name:		PEvalMod
* Desc:		Compute value of polynomial mod Zn
* Stack:	( Q Z Zn --> Q' )
**********************************************************************

NULLNAME PEvalMod
CODE
		GOSBVL	=SAVPTR		First do quick exit for multivariate
		D1=D1+	10
		A=DAT1	A
		D0=A
		GOSUBL	UNIVAR?
		GOC	+
-		GOTO	PEvalModRPL	Multivariate Q - use RPL loop instead

+		GOSBVL	=GETPTR
		GOSUBL	ModPrepare	D0 = ->Q	R4[W] = Zn / ->Zn
		?ST=0	sSHORTMOD
		GOYES	-		Long Zn - use RPL loop instead

* Can do entire calculation in assembly

		C=R4	W		C[W] = Zn
		GOSUBL	ZMODPSHORT
		R3=A	W		R3[W] = Z mod Zn
		D1=D1+	5
		A=DAT1	A
		D0=A			D0 = ->Q
		A=DAT0	A
		LC(5)	=DOLIST
		?A=C	A
		GOYES	+

* Peval(const,x) mod Zn = const mod Zn

		C=R4	W		C[W] = Zn
		GOSUBL	ZMODSHORT
		C=D	S		A[W]C[S] = Q mod Zn
		GOLONG	PushZ3Loop

+		A=0	W
		R0=A	W		R0[W] = result = 0
		D0=D0+	5
PEvalModLoop	A=DAT0	A
		LC(5)	=DOLIST
		?A=C	A
		GOYES	PEvalModRPL	Can't handle multivariate case
		LC(5)	=SEMI
		?A=C	A
		GOYES	pevmodsemi
		A=R0	W		A[W] = result
		C=R4	W
		B=C	W		B[W] = Zn
		C=R3	W		C[W] = Z
		GOSUBL	SHORTMULMOD	A[W] = result*Z mod Zn
		R0=A	W		A[W] = result'
		C=B	W		C[W] = Zn
		GOSUBL	ZMODPSHORT	A[W] = coeff mod Zn
		B=C	W		B[W] = Zn
		C=R0	W		C[W] = result'
		A=A+C	W
		GOC	+
		?A<B	W
		GOYES	++
+		A=A-B	W
++		R0=A	W		R0[W] = result' + coeff mod Zn
		GOTO	PEvalModLoop

pevmodsemi	A=R0	W		A[W] = result
		C=R4	W		C[W] = Zn
		GOSUBL	ZMODSYM		Convert MODP to MOD
		C=D	S		A[W]C[S]=result
		GOLONG	PushZ3Loop

PEvalModRPL	GOSUB	ModRPL:
ENDCODE
* This is reached when Q is multivariate or Zn is long
* The code assumes QMod is always gives extra speed
::
  UNROTSWAP				( Zn Z Q )
  DUPTYPELIST? NOTcase
  :: SWAPDROPSWAP QMod ;		( --> Q Mod Zn )

  PEvalFast?				( Zn Z Q FALSE / Zn Q[Z] TRUE )

  case :: SWAP QMod ;			( --> Q[Z] mod Zn )

  FIVE
  NEXTCOMPOB DROP			( Zn Z Q #off Q[1] )
  5PICK QMod				( Zn Z Q #off Q[1]modZn )
  BEGIN
	4UNROLL				( Zn Qres Z Q #off )
	NEXTCOMPOB			( Zn Qres Z Q #off' Q[i] )
  WHILE
	5ROLL 5PICK			( Zn Z Q #off' Q[i] Qres Z )
	QMul QAdd			( Zn Z Q #off' Qres' )
	5PICK QMod			( Zn Z Q #off' Qres'' )
  REPEAT
  2DROP SWAPDROP			( --> Qres )
;

**********************************************************************
* Name:		QMod
* Desc:		Compute symmetric modulo of polynomial vs given Z.
* Stack:	( Q Zn --> Q' )
* Note:		The result is trimmed if Z <> 0.
*		Output for Z == 0 is the polynomial itself (untouched)
**********************************************************************

NULLNAME QMod

CODE
		GOSUB	ModPrepare	D0 = ->Q	R4[W] = Zn / ->Zn
		?ST=0	sSHORTMOD
		GOYES	+
		A=R4	W		Check Zn=0 separately
		?A#0	W
		GOYES	+
		GOSBVL	=GETPTR
		GOVLNG	=DropLoop	Q MOD 0 = Q

* Now dispatch to various cases given sSHORTMOD and type(Q)

+		A=DAT0	A
		LC(5)	=DOLIST
		?A=C	A
		GOYES	PMODZ

* Integer modulo Zn (or inversion)

		?ST=0	sSHORTMOD	Handle long modulo separately
		GOYES	gorplzmod

		C=R4	W		C[W] = Zn
		GOSUBL	ZMODSHORT	A[W] = Zmod	D[S] = Zmodsign
		C=D	S
		GOLONG	PushZ2Loop

gorplzmod	GOTO	DoRplZMod	Do ZMod in RPL


* Polynomial modulo Zn (or inversion)

PMODZ		?ST=1	sSHORTMOD	Handle short modulo in assembly
		GOYES	+
		A=PC
		LC(5)	(ZmodZ)-(*)	Put ZmodZ on stack
		A=A+C	A
		GOSBVL	=GPPushA
		A=PC
		LC(5)	(=DoPS)-(*)	And call DoPS
		A=A+C	A
		PC=(A)

* Allocate memory for the output polynomial

+		GOSUBL	QSIZE		C[A] = obsize
		GOSBVL	=GETTEMP
		AD0EX
		R0=A	A		R0[A] = ->Q'
		GOSBVL	=D1=DSKTOP
		D1=D1+	5
		C=DAT1	A		C[A] = ->Q
		D0=C			->input
		D1=A			->output
		C=0	A
-		R1=C	A		R1[A] = composite depth
PMODZLOOP	A=DAT0	A
		LC(5)	=DOHSTR
		?A=C	A
		GOYES	dozmodz
		DAT1=A	A		Output DOLIST/SEMI
		D0=D0+	5
		D1=D1+	5
		LC(5)	=DOLIST
		?A#C	A
		GOYES	domodsemi
domodlist	C=R1	A
		C=C+1	A		depth++
		GONC	-
domodsemi	C=R1	A
		C=C-1	A		depth--
		?C#0	A
		GOYES	-

		CD1EX			C[A] = ->Qtail
		GOSUBL	ShrinkQ
		D1=D1+	5
		D=D+1	A
		DAT1=A	A		Overwrite Q with Q'
		GOLONG	DoPTrim		And trim the result

dozmodz		C=R4	W		C[W] = Zn
		GOSUB	ZMODSHORT
		C=D	S		A[W]C[S] = Zinv
		GOSUBL	SHORTOUT
		GONC	PMODZLOOP

DoRplZMod	GOSUB	ModRPL:
ENDCODE
LABEL ZmodZ
::
  SWAPOVER			( M A M )
  QDivRem SWAPDROPDUP		( M R R )
  ROTOVER ZIsNeg?		( R R M sflag )
  ITE QAdd QSub			( R -sign[R]*[M-|R|] )
  ZNMin				( Z )
;

**********************************************************************
* Name:		QAddMod
* Desc:		Polynomial addition modulo Zn.
* Stack:	( Q1 Q2 Zn --> Q' )
**********************************************************************
NULLNAME QAddMod
::
  UNROT QAdd SWAP QMod
;

**********************************************************************
* Name:		QSubMod
* Desc:		Polynomial substraction modulo Zn.
* Stack:	( Q1 Q2 Zn --> Q' )
**********************************************************************
NULLNAME QSubMod
::
  UNROT QSub SWAP QMod
;

**********************************************************************
* Name:		QMulMod
* Desc:		Polynomial multiplication modulo Zn.
* Stack:	( Q1 Q2 Zn --> Q' )
* Notes:	The maximum size of any coefficient in the output
*		of a regular multiplication is
*
*			(min(deg(Q1),deg(Q2))+1)*max(Q1)*max(Q2)
*
*		but is usually of the order max(Q1)*max(Q2).
*		If input is not normalized and the coefficients
*		of the problem are such that we would have big
*		coefficients in the regular product it would
*		make sense to make sure the input is modular
*		before calling QMul.
**********************************************************************
NULLNAME QMulMod
::
  ( Normalize input if it seems unnormalized )
*  3PICK PNMax OVER ZNLT? ?SKIP
*  :: ROTOVER QMod UNROT ;
*  OVER PNMax OVER ZNLT? ?SKIP
*  :: SWAPOVER QMod SWAP ;

  ( Multiply and take modulo )

  UNROT QMul SWAP QMod
;

**********************************************************************
* Name:		QDivMod
* Desc:		Polynomial division modulo Zn.
* Stack:	( Q1 Q2 Zn --> Qquo Qrem )
* Notes:	In regular division the coefficients in the remainder
*		can increase very quickly to tens of digits, thus
*		it is important to normalize the coefficients whenever
*		possible.
**********************************************************************
NULLNAME QDivMod
::
  UNROT PTrim SWAP PTrim SWAP			( Zn Q1 Q2 )

  DUPTYPEHSTR? case				( Zn Q1 Z2 )
  :: 3PICK ModInv				( Zn Q1 1/Z2 )
     ROT QMulMod HXS 1 0			( A' 0 )
  ;

  OVER TYPEHSTR? case				( Zn Z1 P2 )
  :: DROP HXS 1 0 UNROTSWAP QMod ;		( 0 ZmodZn )

  ( Zn P1 P2 )

  ( First handle the trivial case )
  DUPLENCOMP ROT			( Zn B n A )
  DUPLENCOMP ROT			( Zn B A m n )
  #<case
  :: HXS 1 0 ROTDROPSWAP ROT QMod ;	( --> 0 A' )

  ( Try fast division using the monic image as divisor )

  OVER PLCZ				( Zn B A lc )
  DupZIsOne? ITE
     UNROTSWAP				( Zn 1 A B )
     :: 4PICK ModInv			( Zn B A 1/lc )
        DUP4UNROLL			( Zn 1/lc B A 1/lc )
        ROTSWAP 5PICK QMulMod		( Zn 1/lc A B' )
     ;

  ( First try applying the modulo directly in DISTDIV? )
  ( This would avoid coefficient explosion )

  CODE
		GOSBVL	=SAVPTR
		D1=D1+	15
		A=DAT1	A			A[A] = ->Zn
		D0=A
		D0=D0+	5			We require max 15 nibble mod
		A=DAT0	A			so we cannot use SHORTPOP?
		LC(5)	5+15+1			sizefield + mantissa + sign
		?A>C	A
		GOYES	qdvmdfail
		C=A	A
		A=0	W
		C=C-CON	A,7			Zn is always nonzero
		D0=D0+	5
		P=C	0
		A=DAT0	WP			And we can ignore the sign
		P=	0
		D0=(5)	(=IRAM@)-4		Save modulo to IRAMBUFF
		C=DAT0	A
		D0=C
		D0=(4)	#100
		DAT0=A	W
		ST=1	sMODULAR		Use modular math
		GOSBVL	=GETPTR
		GOLONG	DDentry			Use DISTDIV? code
qdvmdfail	GOVLNG	=GPPushFLoop		( --> Zn 1/lc A B' FALSE )
  ENDCODE					( --> Zn 1/lc quo rem TRUE )

  ( If success then fix quotient back and exit )
  case
  :: UNROTSWAP 4ROLL QMulMod SWAP ;	( --> quo' rem )


  ( If failed then QDivRem DISTDIV? would fail too so that )
  ( QDivRem would also operate in SRPL - thus we might as well )
  ( use SRPL and optimize it for modular division )

  ( Regular division algorithm made modular )

  DUPLENCOMP ROT			( Zn 1/lc B n A )
  DUPLENCOMP ROT			( Zn 1/lc B A m n )
  #-+1 5UNROLL				( m-n+1 Zn 1/lc B A )
  5PICK #1+_ONE_DO (DO)			( m-n+1 Zn B C )
	DUP CARCOMP			( m-n+1 Zn 1/lc B C Ck )
	3PICK CARCOMP			( m-n+1 Zn 1/lc B C Ck Bk )
	6PICK QDivMod QIsZero?		( m-n+1 Zn 1/lc B C Qk flag )
	NOTcase ErrUndefRes		( undefined result error )
	6UNROLL CDRCOMP			( Qk m-n+1 Zn 1/lc B C' )
	6PICK 3PICK			( Qk m-n+1 Zn 1/lc B C' Qk B )
	FIVE NEXTCOMPOB 2DROP		( .. Zn 1/lc B C' Qk B #off )
	BEGIN
		NEXTCOMPOB
	WHILE
		4PICK QMul 4UNROLL	( .. Qk*Bi Qk B #off' )
	REPEAT
	SWAPDROP LENCOMP #1-		( Qk m-n+1 Zn 1/lc B C' Meta )
	HXS 1 0 ISTOP-INDEX #1- NDUPN
	top& {}N QSub			( Qk m-n+1 Zn 1/lc B C'' )
	( The next QMod is the crucial one for avoiding )
	( coefficient explosion in the division above. )
	( It is a bit questionable though. )
	4PICK QMod			( Qk m-n+1 Zn 1/lc B C''' )
  LOOP
  SWAPDROP				( Qk m-n+1 Zn 1/lc R )
  THREE psh {}N SWAPDROP		( Zn 1/lc R Q )
  ROT 4ROLL QMulMod			( R Q' )
  SWAPDUP NULLCOMP?			( Q R flag )
  casedrop HXS 1 0			( Q R' )
;

**********************************************************************
* Name:		QInvMod
* Desc:		Polynomial inversion modulo Zn.
* Stack:	( Q Zn --> Q' )
**********************************************************************
NULLNAME QInvMod
::
  DupQIsZero? caseDROP

  ( Handle integer inversion )

  SWAP DUPTYPEHSTR? case
  :: OVER ModInv SWAP QMod ;

  ( Invert each coefficient separately )

  ZERO				( Zn Q #off )
  BEGIN
     NEXTCOMPOB			( ... Zn Q #off C )
  WHILE
     4PICK QInvMod 4UNROLL	( ... C' Zn Q #off )
  REPEAT
  LENCOMP SWAPDROP {}N		( Q' )
;

**********************************************************************
* Name:		QGcdMod
* Desc:		Polynomial GCD modulo Zn for univariate polynomials.
*		The result is made monic.
* Stack:	( Q1 Q2 Zn --> Q' )
**********************************************************************
NULLNAME QGcdMod
::
  UNROT						( Zn A B )
  PTrim SWAP PTrim				( Zn B A )
  DUPTYPELIST? ROT DUPTYPELIST? ROTAND		( Zn A B lists? )
  NOTcase :: 3DROP HXS 2 10 ;			( gcd=1 )

  Univar? NcaseTYPEERR				( * B not univariate* )

  BEGIN
	OVER 4PICK				( Zn A B A Zn )
	QDivMod SWAPDROP			( Zn B C )
	DUPUNROT QIsZero?			( Zn C B flag )
  UNTIL
  SWAPDROP					( Zn G )
  DUPTYPEHSTR? case2drop HXS 2 10		( --> 1 )
  DUP CARCOMP ROT	 			( G lc Zn )
  QDivMod DROP					( Gmonic )
;

**********************************************************************
* Name:		QGcdExMod
* Desc:		Extended polynomial GCD modulo Zn for univariate
*		polynomials.
*
*		The equation: Q1*Q1' + Q2*Q2' = 1  MOD Zn
*
* Stack.	( Q1 Q2 Zn --> Q1' Q2' )
**********************************************************************
NULLNAME QGcdExMod
::
  UNROT					( Zn A B )
  HXS 1 0 HXS 2 10			( Zn A B 1 0 )
  4ROLLOVER 4PICK			( Zn B 1 0 A 0 1 )
  BEGIN					( Zn s1 s2 s3 t1 t2 t3 )
	6ROLL 4PICK 8PICK		( Zn s2 s3 t1 t2 t3 s1 t1 Zn )
	QDivMod				( Zn s2 s3 t1 t2 t3 uq u1 )
	7ROLL 5PICK 4PICK		( Zn s3 t1 t2 t3 uq u1 s2 t2 uq )
	QMul QSub 8PICK QMod		( Zn s3 t1 t2 t3 uq u1 u2 )
	7ROLL 5PICK 5ROLL		( Zn t1 t2 t3 u1 u2 s3 t3 uq )
	QMul QSub 7PICK QMod		( Zn t1 t2 t3 u1 u2 u3 )
	3PICK QIsZero?			( Zn t1 t2 t3 u1 u2 u3 u1=0? )
  UNTIL					( Zn t1 t2 t3 u1 u2 u3 )
  3DROP					( Zn t1 t2 t3 )
  ROT PLCZ				( Zn t2 t3 lc )
  4PICK ModInv				( Zn t2 t3 lc^-1 )
  DUPUNROT 5PICK QMulMod		( Zn t2 lc^-1 t3' )
  UNROT 4ROLL QMulMod SWAP		( t2' t3' )
;


**********************************************************************
*			UTILITIES
**********************************************************************

ASSEMBLE
*---------------------------------------------------------------------
* Modular math preparation
* Stack:	( Q Zn )
* Exit:		sSHORTMOD=1	R4[W]=Zn	D0=->Q
*		sSHORTMOD=0	R4[A]=->Zn	D0=->Q
*---------------------------------------------------------------------

ModPrepare	GOSBVL	=SAVPTR
		A=DAT1	A		->Zn
		D1=D1+	5
		R4=A	A		R4[A] = ->Zn
		ST=0	sSHORTMOD	Assume Zn is long
		D0=A
		GOSUBL	SHORTPOP?	A[W]C[S] = Zn
		GOC	+		It was long - continue
		ST=1	sSHORTMOD	It wasn't - flag so and save Zn
		R4=A	W
+		A=DAT1	A
		D0=A			D0 = ->Q
		RTN

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Desc:		Modular inverse
* Entry:	D0	= ->Z
*		R4[W]	= Zn
* Exit:		A[W]	= Z inv Zn
*		D0	= ->Ztail
* Alters:	A[W] B[W] C[W] D[W] R0[W] R1[W] R2[W] R3[W]
* Stack:	4
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

ZMODINV		C=R4	W
		R3=C	W		R3[W]=P=N
		GOSUB	ZMODPSHORT	A[W] = Zmod	C[W] = Zn
		R2=A	W		R2[W] = Z
		?A=0	W
		RTNYES			ModInv(1,1) = 0
		A=A-1	A
		?A#0	W
		GOYES	+
		A=A+1	A		ModInv(1,any) = 1
		RTN
+		A=A+1	A
		GOSUBL	SHORTGCD	A[W] = GCD(Z,Zn)
		A=A-1	W
		?A=0	W
		GOYES	+
		A=0	W		ModInv(n*x,m*x) = 0
		RTN
+		R1=A	W		R1[W]=S=0
		A=A+1	A
		R0=A	W		R0[W]=R=1

ZMODINVLOOP	A=R2	W		A[W]=Z
		C=R3	W		C[W]=P
		?C=0	W
		GOYES	ZMODINVOK	Until P=0
		GOSUBL	BDIV		A[W]=Q=Z/P
		C=A	W		C[W]=Q
		AR4EX	W		R4[W]=Q
		B=A	W		B[W]=N
		A=R1	W		A[W]=S
		GOSUBL	SHORTMULMOD	A[W]=C[W]=SQ mod N
		C=R0	W		C[W]= R
		C=C-A	W		C[W]=R-SQ mod N
		GONC	+
		C=C+B	W		C[W]=T=R-SQ+N mod N
+		CR1EX			S=T
		R0=C	W		R=S
		A=R3	W		A[W]=P
		C=B	W
		CR4EX	W		C[W]=Q	R4[W]=N
		GOSUBL	SHORTMULMOD	A[W]=C[W]=PQ mod N
		C=R2	W		C[W]=Z
		C=C-A	W		C[W]=Z-PQ mod N
		GONC	+
		C=C+B	W		C[W]=T=Z-PQ+N mod N
+		CR3EX	W		P=T
		R2=C	W		Z=P
		GOTO	ZMODINVLOOP
ZMODINVOK	A=R0	W		A[W]=R
		RTN

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Desc:		Compute symmetric modulo Z1 mod Z2 where Z2 is short
* In:		D0	->Z1
*		C[W]	Z2
* Out:		D0	->Z1tail
*		A[W]	Zmod
*		D[S]	Zmodsign
*		C[W]	Z2
* Alters:	D0 A[W] B[W] C[W] D[S,14,A]
* Stack:	3
* Note:		This always returns the symmetric modulo, not the
*		antisymmetric modulo as the old system RPL code would
*		when Z2 is negative.
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
ZMODSHORT	GOSUB	ZREMSHORT	A[W]=remainder	C[W]=Z2
		GOSUB	ZMODSYM
		?A=0	W		Ignore next if rem is negative
		RTNYES
		B=C	S		Negate modulo of Z1 is negative
		D0=D0-	1
		C=DAT0	S		C[S] = Z1sign
		D0=D0+	1
		CBEX	S
		?B=0	S
		RTNYES
		D=-D-1	S
		RTN

* Little utility to convert MODP to MOD

ZMODSYM		D=0	S		Assume modulo is positive
		?A=0	W
		RTNYES			It is if remainder is zero
		B=A	W		B[W] = remainder
		B=B+B	W
		GOC	+
		?B<=C	W		If 2*rem <= Zn then is symmetric
		RTNYES
+		A=A-C	W
		A=-A	W
		D=-D-1	S
		RTN
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Desc:		Compute positive modulo Z1 mod Z2 where Z2 is short
* In:		D0	->Z1
*		C[W]	Z2
* Out:		D0	->Z1tail
*		A[W]	Zmod
*		D[S]	Zmodsign
*		C[W]	Z2
* Alters:	D0 A[W] B[W] C[W] D[S,14,A]
* Stack:	3
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
ZMODPSHORT	GOSUB	ZREMSHORT	A[W]=remainder	C[W]=Z2
		B=C	S		Negate modulo of Z1 is negative
		D0=D0-	1
		C=DAT0	S		C[S] = Z1sign
		D0=D0+	1
		CBEX	S
		?B=0	S
		RTNYES			rem is result
		?A=0	W		Don't change 0 to Z2 !!!
		RTNYES
		A=A-C	W
		A=-A	W		Z2-rem is result for Z1<0
		RTN

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Desc:		Compute remainder in Z1/Z2 calculation.
* In:		D0	->Z1
*		C[W]	Z2		Z2 is short and unsigned
* Out:		D0	->Z1tail
*		A[W]	Zrem
*		C[W]	Z2
* Alters:	D0 A[W] B[W] C[W] D[S,A]
* Stack:	2
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

ZREMSHORT	B=C	W		B[W] = divisor
		D0=D0+	5
		A=DAT0	A		Z1size

		C=C-1	W
		GOC	z1rem1		Handle Z1 REM 0
		C=C-1	W
		GOC	z1rem1		Handle Z1 REM 0
		C=C-1	W
		GOC	z1rem2		Handle Z1 REM 2

		LC(5)	5+16+1
		?A>C	A
		GOYES	zremlong

* The integer in D0 is short - we can do the entire calculation in registers

		GOSUBL	shortpop	A[W]C[S] = Z1
		CBEX	W		A[W]B[S] = Z1
		GOSUBL	BDIV		C[W]=divisor
		A=B	W		A[W]=remainder
		RTN

z1rem1		CD0EX
		C=C+A	A		Skip Z1
		CD0EX
		C=B	W		C[W] = Z2
		A=0	W		A[W] = remainder = 0
		RTN
z1rem2		D0=D0+	5
		C=DAT0	B		C[0] = low nibble in Z1 (0 for Z0)
		CD0EX
		C=C+A	A		Skip Z1
		CD0EX
		D0=D0-	5		->Z1tail
		A=0	W		A[W] = remainder = 0
		?CBIT=0	0
		GOYES	+
		A=A+1	A		A[W] = remainder = 1
+		C=B	W		C[W] = divisor
		RTN

* The integer in D0 is too long to fit into a register, but the
* remainder can be calculated in registers. (See ZQUO for comparison)
		
zremlong	A=A-CON	A,7		A[A]=nibbles-1
		CD0EX
		C=C+A	A
		CD0EX
		D0=D0-	17-7		->Z1 high 16 nibbles
		C=A	A
		D=C	A		D[A]=nibbles-1
		D=D-CON	A,16		D[A]=nibbles to go
		A=DAT0	W		A[W]=quo1
		C=B	W		C[W]=z2
		GOSUBL	BDIV		B[W]=rem1	C[W]=z2
		A=B	W		A[W]=rem1
		

* Now loop to add the remaining nibbles into the result

zremlonglp	D0=D0-	1		->Z1[i]
		?A#0	S		Handle the bad case of a 64-bit rem
		GOYES	zrembad
		ASL	W
		A=DAT0	1
		?A<C	W
		GOYES	zremlongcont	Avoid BDIV if possible
		GOSUBL	BDIV
		A=B	W		A[W] = rem[i]
zremlongcont	D=D-1	A
		GONC	zremlonglp
		B=C	A		B[A]=Z2[A]
		D0=D0-	5		->Z1size
		C=DAT0	A
		AD0EX
		A=A+C	A		A[A]=->Z1tail
		AD0EX
		C=B	A		C[W]=Z2
		RTN
* Here we need the remainder of Z[17]/Z[16]
zrembad		B=A	S
		A=DAT0	S
		ABEX	S		B[S] = new low nibble
		P=	16-4		Loop 4 times
		
zrembadlp	A=A+A	W
		GOC	zremoverflow	Handle overflow
		B=B+B	S
		GONC	+
		A=A+1	A		New low bit
+		A=A-C	W		Take remainder
		GONC	+
		A=A+C	W		Fix back - was already valid
+		P=P+1
		GONC	zrembadlp
		GOC	zremlongcont

zremoverflow	B=B+B	S
		GONC	+
		A=A+1	A		New low bit
+		A=A-C	W		Always valid here
		P=P+1
		GONC	zrembadlp
		GOC	zremlongcont

RPL
**********************************************************************
