	STITLE	ALG P Factor
**********************************************************************
*		ALG library: Polynomial Factorization Routines
**********************************************************************

**********************************************************************
* EXPORT:
*
*  PFactor	( P --> Lf Z )
*  PSqff	( P --> Lf )
*  PSQFF	( P --> Lsqff )
*
**********************************************************************


**********************************************************************
* Desc:		Do a complete factorization of P
*		The result is trimmed.
*
* Stack:	( P -> Lfk Z )
*
*		where
*
* 		Lfk = { {F1 k1} {F2 k2} ... {Fn kn} }
*		Fi  = non-const factor i
*		ki  = power of factor i
*	        Z   = constant factor
*
**********************************************************************
NULLNAME PFactor
::
  ( * remove integer content * )

  PPZ SWAP				( Z P' )

  ( * full factorization of P * )
 
  TRUE 
  { LAM FullFact? }
  BIND
    PHFctr				( Z Lf )
  ABND  
  SWAP					( Lf Z )
;

**********************************************************************
* Desc:		Square-free & trivial factorization, including 
*		integer content, of P taken positive
*
* Stack:	( P -> Lfk )
*
*		where
*
* 		Lfk = { {F1 k1} {F2 k2} ... {Fn kn} }
*		Fi  = const factor i
*		ki  = power of factor i
*
* Note:		Factors of same power are not necessarily merged or
*		adjacent, but all Fi's are square-free 
**********************************************************************
NULLNAME PSqff
::
  FALSE
  { LAM FullFact? }
  BIND
    PHFctr				( Z Lf )
  ABND  
;

**********************************************************************
* Desc:		Heuristic factorization of polynomial taken positive
*		LAM FullFact? must be bound
*		
*		LAM FullFact? = TRUE   => full factorization
*		LAM FullFact? = FALSE  => square-free & trivial 
*					  factorization only
* Stack:	( P --> Lf )   
*		
*		Lf = { {F1 #k1} ... {Fn #kn} }
* 
**********************************************************************
NULLNAME PHFctr
::
   DUP PSetSign			( |P| )
   DUPTYPELIST? NOTcase PNFctr  ( -> Lf )

   ( * remove trivial factors * )

   PFactTriv SWAP		( Lft P' )

   ( * factor primitive part * )

   PPP SWAP			( Lft PC PP )
   PHFctr1 SWAP			( Lft Lfpp PC )

   ( * done if content is unit * )

   DupZIsOne? casedrop &COMP	( -> Lft&Lfpp )

   ( * factor content and make factors * )
   ( * constant in main variable       * )

   PHFctr			( Lft Lfpp Lfpc )

   FIVE				( Lft Lfpp Lfpc offset )
   BEGIN
     NEXTCOMPOB
   WHILE			( Lft Lfpp Lfpc offset pi )
     INCOMPDROP SWAP		( Lft Lfpp Lfpc offset #ki fi )
     ONE CKMkP SWAP		( Lft Lfpp Lfpc offset fi' ki )
     TWO{}N UNROT		( Lft Lfpp pi' Lfpc offset )
   REPEAT			( Lft Lfpp p1'..pn' Lfpc )
   LENCOMP {}N			( Lft Lfpp Lfpc' )

   ( * combine all factors * )

   SWAP &COMP &COMP		( Lf )
;

**********************************************************************
* Desc:		Heuristic factorization of primitive polynomial
*		LAM FullFact? must be bound
*		
*		LAM FullFact? = TRUE   => full factorization
*		LAM FullFact? = FALSE  => square-free & trivial 
*					  factorization only
* Stack:	( P --> Lf )   
*		
*		Lf = {} 			if P is constant
*		Lf = { {F1 #k1} ... {Fn #kn} }  o/w
* 
**********************************************************************
NULLNAME PHFctr1
::
  DUPTYPELIST? NOTcasedrop NULL{}	( -> {} )

  ( * square-free factorization * )

  PSQFF					( Lsqf )

  ( * factor non trivial square-free factors * )

  ZEROSWAP				( Meta Lsqf )
  toLEN_DO (DO) 
    ::
      DUPINDEX@ NTHCOMPDROP		( Meta Lsqf Fi )
      DUPTYPELIST? NOTcaseDROP

      LAM FullFact? ITE PHFctr0 ONE{}N	( Meta Lsqf Lfi )

      ( * add factors to Meta * )
      
      FIVE				
      BEGIN				( Meta Lfsq Lfi offset )
        NEXTCOMPOB
      WHILE				( Meta Lfsq Lfi offset fij )
        INDEX@ TWO{}N			( Meta Lfsq Lfi offset {fij i} )
        5UNROLL 4ROLL #1+ 4UNROLL	( Meta' Lfsq Lfi offset )
      REPEAT
      DROP				( Meta Lfsq )
    ;
  LOOP
  DROP {}N				( Lf )  
;

**********************************************************************
* Desc:		Heuristic factorization of primitive square-free 
*		non constant polynomial
* Stack:	( P --> Lf ) 
*		
*		Lf = { F1 .. Fm }
**********************************************************************
NULLNAME PHFctr0
::
  DUPLENCOMP THREE #<case ONE{}N	( deg<2 -> { P } )

  ( * berlekamp if univariate * )

  Univar? case UFactor			( -> Lf )

  ( * Multivariate -> use Z-adic * )

  DUP PNMax ZAbs			( P pmax )		
  HXS 2 20 QMul				( P 2*pmax )
  HXS 2 20 QAdd				( P z )
  2DUP RESOROMP PEval DoLS		( P z Pz )
  PHFctr				( P z Lfz )
  DeCntMulti				( P z Lfz' )
  LiftZAdic				( Lf )
;

**********************************************************************
* Desc:		Transform list with count into simple list
* Stack:	( R -> L )  
* Note:				
*		R = { {f1 #k1} ... {fn #kn} }
*		L = { f1 f1 .. fn fn }
**********************************************************************
NULLNAME DeCntMulti
::
  ZEROSWAP				( #0 R )
  toLEN_DO (DO)				( f1..fm #m R )
     DUPINDEX@ NTHCOMPDROP		( f1..fm #m R ri )
     INCOMPDROP				( f1..fm #m R fi #k )
     NDUPN				( f1..fm #m R fi1 fik #k )
     DUP #3+ ROLL			( f1..fm R fi1 fik #k #m )
     OVER #+SWAP 			( f1..fm R fi1 fik #m' #k )
     #2+ROLL				( f1..fm fi1 fik #m' R )
  LOOP
  DROP {}N
;

**********************************************************************
* Desc:		Apply program F(Li,S) to every elem of L
* Stack:	( L S F --> L' )
**********************************************************************
NULLNAME DoLS
::
  1LAMBIND			( L S )
  SWAP FIVE			( S L offset )
  BEGIN
	NEXTCOMPOB
  WHILE				( .. S L offset' Li )
	4PICK			( .. S L offset' Li S )
	1GETLAM EVAL		( .. S L offset' Li' )
	4UNROLL			( .. Li' S L offset' )
  REPEAT
  ABND				( L1' .. Ln' S L )
  SWAPDROP			( L1' .. Ln' L )
  LENCOMP {}N			( L' )
;

**********************************************************************
* Desc:		Factorization of positive integer as polynomial
* Stack:	( Z -> Lf )  
* Note:				
*		Lf = {} 			if Z is 1
*		Lf = { {Z1 #k1} ... {Zn #kn} }  o/w
**********************************************************************
NULLNAME PNFctr
::
  DupQIsZero? caseSIZEERR
  DupZIsOne? casedrop NULL{}		( -> {} )

  NFactorSpc				( L )
  ZEROSWAP				( #0 L )
  INNERCOMP				( #0 l1..ln #n )
  BEGIN					( Meta l1..ln #n )
    #1-SWAP ONE				( Meta l1..li #i z 1 )
    BEGIN				( * count duplicates * )
      OVER5PICK EQUAL			( Meta l1..li #i z 1 flag )
    WHILE				( Meta l1..li-1 z #i z #k )
      SWAPDROPSWAP			( Meta l1..li-1 z #k #i )
      #1-UNROT #1+			( Meta l1..li-1 #i-1 z #k+1 )
    REPEAT				( Meta l1..li #i z #k )
    TWO{}N				( Meta l1..li #i f={z #k} )
    psh1&rev				( Meta&f l1..li #i )
  #0=UNTIL (UNTIL)			( Meta #0 )
  DROP {}N				( Lf )
;	

**********************************************************************
* Desc:		Compute the square-free factorization of primitive P
* 		The result is trimmed (provided P was)
* Stack:	( P --> Lsqff )
**********************************************************************
NULLNAME PSQFF
::
  ONESWAP			( #1 P )
  DUP PDer			( #1 P P' )
  OVER QGcd			( #1 P R1 )
  SWAPOVER			( #1 R1 P R1 )
  QDiv SWAP			( #1 T1 R1 )
  BEGIN				( S1..Si-1 i Ti Ri )
	DUPTYPELIST?		( S1..Si-1 i Ti Ri flag )
  WHILE
  	ROT#1+UNROT		( S1..Si-1 i+1 Ti Ri )
	2DUP QGcd		( S1..Si-1 i+1 Ti Ri Ti+1 )
	SWAPOVER QDiv		( S1..Si-1 i+1 Ti Ti+1 Ri+1 )
	ROT 3PICK		( S1..Si-1 i+1 Ti+1 Ri+1 Ti Ti+1 )
	QDiv			( S1..Si-1 i+1 Ti+1 Ri+1 Si )
	4UNROLL			( S1..Si-1 Si i+1 Ti+1 Ri+1 )
  REPEAT			( S1..Se-1 e Te Re )
  DROPSWAP {}N			( Le )
;

**********************************************************************
* desc:		lift n-1 z-adic factorization into n factorization
* stack:	( p z F --> L )  
*		
*		F = { f1 .. fm }
*		L = { f1' .. fn' }
**********************************************************************
NULLNAME LiftZAdic
::				( P z F )
  DUPLENCOMP			( P z F r )

  DUPTWO #< case2drop		( * done if r=1 * )
   :: DROP ONE{}N ;		( -> {P} )
	
  NULL{}			( P z F r L )

  {{ P z F r L }}

  r CombInit			( C )

  BEGIN
   ::	
      F LFCProd			( C Qz )
      DUPTYPELIST?		( C Qz flag )
      NOTcasedrop CombNext	( repeat with Cnext )
      z RESOROMP PZadic DoLS	( C Q )
      P PLCZ			( C Q lp )
      OVER PLCZ			( C Q lp lq )
      QDivRem SWAPDROP		( C Q rem[lp,lq] )
      QIsZero?			( C Q flag )
      NOTcasedrop CombNext	( repeat with Cnext )
      P OVER			( C Q P Q )
      QDiv?			( C Q [P' FALSE / TRUE] )
      NOTcasedrop CombNext	( repeat with Cnext )
      PTrim P!			( C Q )
      DUP PSetSign		( C |Q| )
      L SWAP >TCOMP L!		( C )
      F SWAP			( F C )
      RmCombNext		( Frm F' C' flag )
      ROT F!			( Frm C' flag )
      ROTDROP			( C' flag )
      caseTRUE			( -> C' TRUE )
      DROPFALSE                 ( -> FALSE )
   ;
  NOT_UNTIL (UNTIL)
  L P				( L P )
  DUP PSetSign			( L |P| )
  >TCOMP			( L' )
  ABND
;
  
*---------------------------------------------------------------------
* Desc:		Calculate combination product
* Stack:	( C L --> C P )
*---------------------------------------------------------------------
NULLNAME LFCProd
::
  HXS 2 10				( C L 1 )
  3PICK LENCOMP #1+_ONE_DO (DO)		( C L P )
    ::
      3PICK INDEX@ NTHCOMPDROP		( C L P ci )
      #0=?SEMI				( * fi not in this combination * )
      OVER INDEX@ NTHCOMPDROP		( C L P fi )
      QMul				( C L P' )
    ;
  LOOP
  SWAPDROP				( C P )
;
	
**********************************************************************
* Name:		UFactor
* Desc:		Factorization of a square free primitive
*		univariate polynomial.
* Stack:	( P --> Lf )
* Notes:	P	remaining polynomial to be factored
*		Lfp	list of modular factors
*		p	prime
*		r	number of modular factors
*		Lf	true factors
*		bf	bound on factor coefficients
*		pj	next modulo after lifting
*		po	current modulo in lifting
*		q	help variable in lifting
*		t	help variable in lifting
*		lc	leading coefficient
*		Pd	derivative of polynomial
*		Pp	polynomial mod p
*		Pdp	derivative of polynomial mod p
*
**********************************************************************
NULLNAME UFactor
::

* Handle small degree polynomials separately

   DUPLENCOMP					( P #deg+1 )
   DUP #<3 casedrop ONE{}N			( --> Lf )
   #3= case UFactorDeg2				( --> Lf )

* If trailing coefficient is smaller than leading one then reverse
* the coefficients and thus the coefficients of the factors.

   DUPDUP LENCOMP NTHCOMPDROP			( P tc )
   OVER CARCOMP					( P tc lc )
   ZNLT? DUPUNROT				( flag P flag )
   IT :: INNERCOMP reversym {}N ;		( flag P' )

   UFactor1					( flag Lf )

* Abort if no reversal was done

   SWAP NOT?SEMI

* Reverse coefficients of each factor

   INNERCOMP			( f1 .. fn #n )
   DUP ZERO_DO (DO)		( f1 .. fn #n )
       DUP #1+ROLL		( .. fn #n f1 )
       INNERCOMP reversym {}N	( .. fn #n f1' )
       DUP PSetSign		( .. fn #n |f1'| )
       SWAP			( .. fn f1 #n )
   LOOP
   {}N				( Lf' )

;

*---------------------------------------------------------------------
* Desc:		Factorization of a square free primitive
*		univariate polynomial of degree > 2.
* Stack:	( P --> Lf )
*---------------------------------------------------------------------
NULLNAME UFactor1
::

*  "Berlekamp.." DISPROW1

   Berlekamp					( P FALSE / P Lfp #p TRUE )

* Abort if Berlekamp failed

   NOTcase ONE{}N				( --> Lf )

* Test if irreducible mod chosen prime

   OVER LENCOMP #1= case2drop ONE{}N		( --> Lf )

* Prepare lifting of modular factors

   #>Z OVER LENCOMP				( P Lfp p r )
   NULL{} NINE NDUPN
   {{ P Lfp p r Lf bf pj po q t lc Pd Pp Pdp }}

*  "Berle: " r #>$ &$ " factors" &$ DISPROW1
*  "Modulo: " p Z>S &$ DISPROW2

* Make factors monic

   Lfp p MonicLf Lfp!

* Try lifting linear factors first

   LiftLinear

* We're done if
*     1) Less than 2 modular factors left
*     2) 2 modular factors left, first one of which is linear

   r #0=case :: Lf ABND ;
   r #1=
   r #2= Lfp CARCOMP LENCOMP #2= AND
   ORcase    :: Lf P >TCOMP ABND ;

*  "True factors: " Lf LENCOMP #>$ &$ DISPROW4

* Convert from monic to true modular factorization

   Lfp P CARCOMP p DemonicLf Lfp!

* General univariate lifting

   LiftGeneral

* Combine modular factors to true factors

   P Lfp po CombineFac				( Tf Tfp )

* Combine to true factors found earlier

   Lf SWAPDROPSWAP &COMP			( Lf&Tf )

   ABND
;

*---------------------------------------------------------------------
* Desc:		Convert true modular factorization to monic factorization
*		by dividing by the leading coefficient of factor 1
* Stack:	( Lfp p --> Lfp' )
*---------------------------------------------------------------------
NULLNAME MonicLf
::
  SWAP INNERCOMP get1			( u1 u2 .. un #n p )
  OVER #2+ROLL				( u2 .. un #n p u1 )
  DUP PLCZ ROT				( u2 .. un #n u1 lc1 p )
  QDivMod DROP				( u2 .. un #n u1' )
  OVER #1+UNROLL			( u1' .. un #n )
  {}N					( Lfp' )
;

*---------------------------------------------------------------------
* Desc:		Convert monic modular factorization to true modular
*		factorization by multiplying factor1 by lcoeff
* Stack:	( Lfp lc p --> Lfp' )
*---------------------------------------------------------------------
NULLNAME DemonicLf
::
  DUPUNROT QMod				( Lfp p lc' )
  DupZIsOne? case2DROP			( --> Lfp )
  3PICK CARCOMP				( Lfp p lc' u1 )
  UNROTSWAP QMulMod			( Lfp u1' )
  SWAP INNERCOMP			( u1' u1 .. un #n )
  DUP #1+ROLL DROP			( u1' .. un #n )
  {}N					( Lfp' )
;

*---------------------------------------------------------------------
* Desc:		Lift modular roots of a polynomial to find
*		linear factors of a univariate polynomial
* Stack:	( #root1 .. #rootn #n  --> )
* Notes:
*		Lflin   = list of found true factors
*		Lfplin' = remaining linear factors
*		P'      = remaining polynomial
*
*		Assumes UFactor lambda variables available and uses
*		them for input & output.
*
* Algorithm:
*
*		IPRZL algorithm by R. Loos
*		"Rational zeros of integral polynomials by p-adic extensions"
*		SIAM J. of Computing 12, pp 286-293 (1983)
*---------------------------------------------------------------------
NULLNAME LiftLinear
::

* Separate linear and nonlinear modular factors.

  ZERO
  Lfp ZERO NEXTCOMPOB DROP			( #0 Lfp #off u1 )
  BEGIN						( r1 .. #m-1 Lfp #off un )
	DUPLENCOMP #2= ITE
	::
           INNERCOMP DROPSWAPDROP QNeg		( r1 .. #m-1 Lfp #off rm )
	   4UNROLL ROT#1+UNROT			( r1 .. rm #m Lfp #off )
	   NEXTCOMPOB NOT
	;
	:: 2DROP TRUE ;				( r1 .. #m-1 Lfp TRUE )
  UNTIL

* Put nonlinear factors back in Lfp

  OVER #1+ MINUSONE SUBCOMP Lfp!		( r1 .. #m-1 )

*  "Linear factors: " OVER #>$ &$ DISPROW3

* Lift the roots until true linear factors can be tested

  DUP#0=csDROP					( Nothing to lift )

  P PDer Pd!					( Save derivative of P )
  P CARCOMP lc!					( Save lcoeff[P] )

* Lifting bound = 2*lcoeff[P]*tcoeff[P]

  P DUPLENCOMP NTHCOMPDROP			( tcoeff[P] )
  lc QMul HXS 2 20 QMul ZAbs			( 2*lc*tc )
  bf!

* Lift roots above bound

  p po!						( po=p )
  BEGIN
	po bf ZNLT?				( po<bf ? )
  WHILE
	po DUP QMul pj!				( pj=po^2 )
	P pj QMod Pp!				( Pp=P mod pj )
	Pd po QMod Pdp!				( Pdp=Pd mod po )
	DUP ZERO_DO (DO)
	    DUP #1+ROLL				( .. root )
	    Pp OVER pj PEvalMod			( .. root Pp[root] )
	    po QDiv				( * always divisible * )
	    Pdp 3PICK po PEvalMod		( .. root Pp[root] Pdp[root] )
	    po ModDiv				( .. root adjust )
	    po QMul QSub			( .. root' )
	    SWAP
	LOOP
	pj po!					( po=pj )
  REPEAT

* Convert roots to polynomials in integer domain and do trial division
* Any true factors found are saved in lams and P is updated, any
* false factors are restored back to original modulo and kept on stack.

  DUP ZERO_DO (DO)
	DUP #1+ROLL				( .. root )
	lc po QMod				( .. root lc' )
	DUP3PICK po QMulMod QNeg		( .. root lc' tc' )
	TWO{}N PPZ DROP				( .. root f )
	P OVER QDiv? ITE
	::					( * Found true factor * )
	  P!					( Update P )
	  Lf SWAP >TCOMP Lf!			( Update Lf )
	  DROP#1-				( drop the root )
	;
	::					( * Not true factor * )
	  HXS 2 10 SWAPDROPSWAP			( .. 1 root )
	  QNeg p QMod TWO{}N			( .. fp )
	  SWAP					( save the modular factor )
	;
  LOOP

* Put back the remaining linear factors to the list of modular factors

  Lfp INNERCOMP top&
  DUP r!
  {}N Lfp!
;

*---------------------------------------------------------------------
* Desc:		Lift factorization mod p to factorization mod p^k
*		where p^k exceeds the factor bound for succesful
*		true factor extraction.
* Stack:	( --> )
* Notes:	Assumes UFactor lambda variables
*---------------------------------------------------------------------
NULLNAME LiftGeneral
::
  ( Init po here in case we get an early abort )
  p po!

  ( Calculate bound = 2*sqrt[lc]*H+2 )
  P PNMax ZAbs HXS 2 20 QMul		( 2*H )
* P CARCOMP ZAbs ZSQRT DROP QMul	( 2*H*sqrt[lc] )
  HXS 2 40 QAdd				( 2*H*sqrt[lc]+4 )
  bf!

* "Lifting " r #>$ &$ DISPROW5
* "Ht " bf Z>S &$ DISPROW6

  ( Abort if no lifting is needed )

  bf p ZNLT? ?SEMI

  ( Partial fraction expansion of 1/P )

* "Partial fractions" DISPROW7
  Lfp INNERCOMP					( u1 .. un #n )
  HXS 2 10 SWAP ONE_DO (DO)			( wi-1 .. u1 .. un-i )
      p QMulMod					( wi-1 .. u1 .. wi )
      DUP ISTOP-INDEX #2+UNROLL			( wi wi-1 .. u1 .. )
  LOOP
  2DROP						( wn-1 .. w2 w1 )

  HXS 2 10 q!					( q=1 )
  r ONE_DO (DO)
      Lfp INDEX@ NTHCOMPDROP
      2DUPSWAP p QGcdExMod
      q p QMulMod
      ROT p QDivMod SWAPDROP UNROT
      q p QMulMod
      SWAP p QDivMod q! DROP
      r #1- UNROLL
  LOOP
  q r UNROLL					( vn .. v1 )

  ( Linear lifting loop )

  p DUP QMul pj!				( pj=p^2 )

  BEGIN
*     "pj " pj Z>S &$ DISPROW7

      ( Update lcoeff[u1] to be lcoeff[P] mod pj )

      Lfp INNERCOMP				( u1 .. un #n )
      Lfp CARCOMP INNERCOMP			( u1 .. un #n Meta )
      P CARCOMP pj QMod				( u1 .. un #n Meta lc' )
      OVER #1+UNROLL				( u1 .. un #n lc Meta' )
      {}N SWAPDROP				( u1 .. un #n u1' )
      OVER #1+UNROLL				( u1 u1' .. un #n )
      {}N Lfp! DROP				(  )

      ( Calculate t = P-u1*..*un mod pj / po )

      P						( P )
      Lfp INNERCOMP ONE_DO (DO)
         pj QMulMod				( P u1 .. un-i*..*un )
      LOOP
      pj QSubMod				( P-u1*..*un )
      po QDiv t!

      ( Lift the factors )
      ( If t is zero then all adjustments are zero?? )

      r #1+_ONE_DO (DO)
*            "Lift u[" INDEX@ #>$ &$ "]" &$ DISPROW7
             Lfp INDEX@ NTHCOMPDROP		( ui )
             INDEX@ #2* PICK			( ui vi )
             t p QMulMod			( ui vi*t_p )
             OVER p QDivMod SWAPDROP		( ui rem[ui,vi*t] )
             po QMul				( ui po*rem[ui,vi*t] )
             pj QAddMod				( ui' )
      LOOP

      r {}N Lfp!				( Update list of factors )

      pj po!					( po=pj )
      pj p QMul pj!				( pj=pj*p )

      bf po ZNLT?				( bf<po ? )

  UNTIL

  ( Drop vi from stack )

  r NDROP
;


*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		UFactorDeg2
* Desc:		Factorization of a degree 2 polynomial, polynomial
*		is univariate, square free and primitive
* Stack:	( P --> Lf )
* Notes:	P = AX^2+BX+C
*		==> Discriminant = B^2-4AC = D^2
*		==> If not square then cannot factorize
*		Else f1 = X-(-B+D)/2A
*		     f2 = X-(-B-D)/2A
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NULLNAME UFactorDeg2  
::
  DUP INCOMPDROP				( P a b c )
  OVERDUP QMul					( P a b c b^2 )
  SWAP4PICK QMul HXS 2 40 QMul QSub		( P a b b^2-4ac )
  DupZIsNeg? ITE FALSE ZSQRT			( P a b d flag )
  NOTcase :: 3DROP ONE{}N ;			( --> Lf )
  ROT HXS 2 20 QMul				( P b d 2a )
  DUP 4PICK 4PICK QSub				( P b d 2a 2a b-d )
  TWO{}N PPZ DROP				( P b d 2a f1 )
  4UNROLL					( P f1 b d 2a )
  UNROT QAdd					( P f1 2a b+d )
  TWO{}N PPZ DROP				( P f1 f2 )
  TWO{}N SWAPDROP				( f1 f2 )
;

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		CombineFac
* Desc:		Combine modular factors to true factors
* Stack:	( P Lfp p --> Tf Tfp )
* Notes:
*		P	polynomial to factor
*		Lfp	list of modular factors
*		p	modulo
*		r	number of modular factors
*		lc	leading coefficient of P
*		Cb	combination list
*		Tf	list of found true factors
*		Tfp	list of modular factors for each true factor
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NULLNAME CombineFac
::

* Make factors monic

  DUPUNROT MonicLf SWAP			( P Lfp p )

* Check input size

  OVER LENCOMP				( P Lfp p r )
  DUPTWO #< case2drop			( Done if r=1 )
  :: SWAP ONE{}N SWAP ONE{}N ;		( --> {P} {Lfp} )

  4PICK PLCZ				( P Lfp p r lc )
  NULL{} DUPDUP				( P Lfp p r lc Tf Tfp Cb )
  8PICK HXS 2 10 PEval			( P Lfp p r lc Tf Tfp Cb P[1] )

  {{ P Lfp p r lc Tf Tfp Cb v1 }}

* Initialize combination to { 1 0 0 0 .. }

  r CombInit Cb!

* Counter for debugging purposes

* ZERO

* Combine until true factors found. The product of a subset of modular
* factors is guaranteed to be a polynomial since the leading coefficient of
* the polynomial being factored is not divisible by p.
* Experimentally divisibility of P(1) by F(1) is an excellent predictor
* on whether F divides P, thus we use it instead of the more common
* test on leading/trailing coefficients.

  BEGIN
*	#1+ "Combination " OVER #>$ &$ DISPROW7

	lc Lfp p Cb				( lc Lfp p Cb )
	CombProd				( F )
	v1 OVER HXS 2 10 PEval			( F P[1] F[1] )
	Mod QIsZero?				( F flag )
	ITE
	  :: P OVER QDiv? ;			( F quo TRUE / F FALSE )
	  FALSE
	ITE
	::					( * Found true factor * )
	  DUP PLCZ lc! P!			( Update lc and P )
	  Tf SWAP >TCOMP Tf!			( Update true factors )
	  Lfp Cb RmCombNext			( Update factors & comblist )
          4UNROLL				( flag Lfrm Lfp' Cb' )
          Cb! Lfp!
          Tfp SWAP >TCOMP Tfp!			( flag )
          NOT
	;
	::
	  DROP					( Drop F )
	  Cb CombNext				( Next combination )
	  NOTcaseTRUE				( --> TRUE )
	  Cb! FALSE				( --> FALSE )
	;
  UNTIL

* Drop the debugging counter

* DROP

* The remaining polynomial is irreducible

  Tf P >TCOMP					( Tf&P )
  Tfp Lfp >TCOMP				( Tf&P Tfp&Lfp )

  ABND

;


*---------------------------------------------------------------------
* Desc:		Calculate modular combination
* Stack:	( lc Lfp p Cb --> F )
*---------------------------------------------------------------------
NULLNAME CombProd
::
  HXS 2 10				( lc Lfp p Cb 1 )
  OVER LENCOMP #1+_ONE_DO (DO)		( lc Lfp p Cb P )
    ::
      OVER INDEX@ NTHCOMPDROP		( lc Lfp p Cb P ci )
      #0=?SEMI				( * fi not in this combination * )
      4PICK INDEX@ NTHCOMPDROP		( lc Lfp p Cb P fi )
      4PICK QMulMod			( lc Lfp p Cb P' )
    ;
  LOOP
  SWAPDROP ROTDROP			( lc p P' )
  UNROT QMulMod				( lc*P' )
  PPZ DROP				( F )
;

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Combination list is a list of ONE's and ZERO's where ONE indicates
* the corresponding element in the selection is to be used in this
* combination. Mirror images are never generated.
*
*	init:		{ 1 0 0 0 0 0 ... 0 0 0 }
*	next:		{ 0 1 0 0 0 0 ... 0 0 0 }
*	next:		{ 0 0 1 0 0 0 ... 0 0 0 }
*	...
*	next:		{ 0 0 0 0 0 0 ... 0 0 1 }
*	next:		{ 1 1 0 0 0 0 ... 0 0 0 }
*	next:		{ 1 0 1 0 0 0 ... 0 0 0 }
*	...
*	next:		{ 0 0 0 0 0 0 ... 1 1 1 } half ones
*	next:		-
*
* Next combination will not be generated if there would be more
* ONE's than ZERO's, or there are an equal number and all the ONE's
* are at the end of the list.
*
* Removing a combination is easy, one merely reduces the list size
* by the number of ONE's, keeps the first one where it is and adds
* the remaining ONE's right after it. If this would overflow the
* list the number of ONE's is increased and they are written to
* the start of the list instead. The same limitations on the number
* and position of the ONE's apply now with the reduced list size.
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


*---------------------------------------------------------------------
* Desc:		Init modular combination list to value { 1 0 0 0 .. }
* Stack:	( #r --> Cb )
*---------------------------------------------------------------------
NULLNAME CombInit
::
  ONE ZERO ROT #1- NDUPN #1+ {}N
;

*---------------------------------------------------------------------
* Desc:		Get next possible modular combination
* Stack:	( Cb --> Cb' TRUE / FALSE )
* Notes:	Assumes Cb is valid and is in tempob area
*---------------------------------------------------------------------
NULLNAME CombNext
CODE
		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A			D0 = ->Cb

* Skip to end of Cb and count elements at the same time

		D0=D0+	5	
		LC(5)	=SEMI
		B=0	A
		B=B-1	A
-		B=B+1	A		elements++
		A=DAT0	A
		D0=D0+	5
		?A#C	A
		GOYES	-

* Count trailing ONE's

		D0=D0-	5
		LC(5)	=ZERO
		D=0	A
		D=D-1	A
-		D=D+1	A		ones++
		D0=D0-	5
		A=DAT0	A
		DAT0=C	A		replace them all with ZERO's
		?A#C	A
		GOYES	-

* Skip back over all zeros

-		D0=D0-	5
		A=DAT0	A
		?A=C	A
		GOYES	-

		LC(5)	=DOLIST
		?A=C	A
		GOYES	+		Skip if all ones were at the end
		LC(5)	=ZERO
		DAT0=C	A		Else write ZERO over ONE
+		C=D	A
		A=C	A		A[A] = ones
		A=A+1	A
		A=A+A	A
		?A>B	A		If we would have more ONE's than
		GOYES	combnextfail	zeros then we're done

		LC(5)	=ONE		Write the new ONE's
-		D0=D0+	5
		DAT0=C	A
		D=D-1	A
		GONC	-

		?A<B	A		If we have less ones than zeros
		GOYES	combnextok	then we're done

		D0=D0+	5		Else ones=zeros and we must disallow
		A=DAT0	A		the case when all ones are at the
		LC(5)	=SEMI		end of the list
		?A=C	A
		GOYES	combnextfail

combnextok	GOVLNG	=GPPushTLoop
combnextfail	GOVLNG	=GPOverWrFLp

ENDCODE

*---------------------------------------------------------------------
* Desc:		Remove next possible combination after a successful
*		combination has been found, and remove the used
*		factors from the factor list
* Stack:	( Lf Cb --> Lfrm Lf' Cb' flag )
*---------------------------------------------------------------------
NULLNAME RmCombNext
::

* Split Lf to a list of removed factors and a list of remaining ones

  NULL{} NULL{} 2SWAP			( Lfrm Lf' Lf Cb )
  toLEN_DO (DO)
     OVER INDEX@ NTHCOMPDROP		( Lfrm Lf' Lf Cb Lfi )
     OVER INDEX@ NTHCOMPDROP		( Lfrm Lf' Lf Cb Lfi Cbi )
     #0=ITE
     :: 4ROLLSWAP >TCOMP UNROT ;	( Lfrm Lf'&Lfi Lf Cb )
     :: 5ROLL SWAP >TCOMP 4UNROLL ;	( Lfrm&Lfi Lf' Lf Cb )
  LOOP
  SWAPDROP				( Lfrm Lf' Cb )

  CODE
		GOSBVL	=PopASavptr
		D0=A			D0 = ->Cb

* Count the number of elements, number of ONE's and position of first ONE

		D0=D0+	5
		B=0	A		elements=0
		D=0	A		ones=0
		C=0	A
		R3=C	A		pos1=0
-		B=B+1	A		elements++
		A=DAT0	A
		D0=D0+	5
		LC(5)	=SEMI
		?A=C	A
		GOYES	++		Done counting
		LC(5)	=ZERO
		?A=C	A
		GOYES	-		Done updating if zero
		D=D+1	A		ones++
		C=R3	A		C[A] = pos1
		?C#0	A
		GOYES	-		Already found 1st one - continue
		C=B	A
		R3=C	A		pos1 = elements
		GONC	-
++

		A=B	A		A[A] = elements+1
		C=D	A
		R2=C	A		R2[A] = ones

* New size = elements-ones

		A=A-C	A
		A=A-1	A
		R1=A	A		R1[A] = elements (new)

* If pos1+ones-1>elements then we must increase number of ones

		A=R3	A
		A=A+C	A		A[A] = pos1+ones
		A=A-1	A		
		C=R1	A		A[A] = elements
		?A<=C	A
		GOYES	+
		C=R2	A
		C=C+1	A		ones++
		R2=C	A
		C=0	A
		C=C+1	A
		R3=C	A		pos1=1
+

* Next combination not possible if
*	2*ones>elements
*	2*ones=elements & pos1>ones

		A=R1	A		A[A] = elements
		C=R2	A
		C=C+C	A		C[A] = 2*ones
		?C>A	A
		GOYES	rmcbnextfail
		?C#A	A
		GOYES	+		2*ones<elements - continue
		A=R3	A
		A=A+A	A		A[A] = 2*pos1
		?A<=C	A
		GOYES	+		pos1<=ones - continue

rmcbnextfail	GOVLNG	=GPPushFLoop	( --> FALSE )

* Allocate memory for next combination and create it

+		A=R1	A		A[A] = elements
		A=A+1	A
		A=A+1	A		+2 to get DOLIST - SEMI
		C=A	A
		C=C+C	A
		C=C+C	A
		C=C+A	A		C[A] = 2*(elements+2)
		GOSBVL	=GETTEMP	D0=->memory	D1=->memorytail

		AD0EX
		R0=A	A		R0[A] = ->Cb (new)

		LC(5)	=SEMI		Terminate Cb
		D1=D1-	5
		DAT1=C	A
		D1=D1-	5
		LC(5)	=ZERO
		A=R1	A		A[A] = elements
-		DAT1=C	A		Initialize Cb to ZERO's
		D1=D1-	5
		A=A-1	A
		?A#0	A
		GOYES	-
		LC(5)	=DOLIST
		DAT1=C	A		Start Cb

		C=R3	A		Skip to pos1 position
		AD1EX
		A=A+C	A
		C=C+C	A
		C=C+C	A
		A=A+C	A
		AD1EX

		A=R2	A		A[A] = ones
		LC(5)	=ONE
-		DAT1=C	A		Write the ONE's
		D1=D1+	5
		A=A-1	A
		?A#0	A
		GOYES	-

		A=R0	A
		GOSBVL	=GPPushA
		GOVLNG	=PushTLoop	( --> Cb' TRUE )
  ENDCODE
  caseTRUE				( --> Lfrm Lf' Cb' TRUE )
  NULL{} FALSE				( --> Lfrm Lf' {} FALSE )
;

**********************************************************************
* Desc:		Extract all trivial power factors of P
* Stack:	( P --> P' Lf )
*		Lf = { {F1 #k1} .. {Fj #kj} }
**********************************************************************
NULLNAME PFactTriv
::
  PFactPowCnt				( P Lk flag )
  NOTcasedrop NULL{}			( --> P {} )
  SWAPOVER PDivLk 			( Lk P' )
  ZERO ROT				( P' #0 Lk )
  toLEN_DO (DO)				( P' f1 .. fj #j Lk )
    ::
       DUPINDEX@ NTHCOMPDROP		( P' f1 .. fj #j Lk #k )
       DUP#0=csDROP			( * continue if 0 * )
       INDEX@ ONE MkSplP		( P' f1 .. fj #j Lk #k Fi )
       SWAP TWO{}N			( P' f1 .. fj #j Lk fi )
       UNROT SWAP#1+SWAP		( P' f1 .. fj fi #j+1 Lk )
    ;
  LOOP 
  DROP {}N
;

**********************************************************************
* Desc:		Calculate trivial power factors in P
* Stack:	( P --> P Lk flag )
*		Lk = { #k1 .. #kn }
* Notes:	flag is TRUE if any of the powers is nonzero
**********************************************************************
NULLNAME PFactPowCnt
CODE

sCNTZERO	EQU 9	* Flag to indicate a nonzero trivial factor

		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A			D0 = ->P
		GOSUBL	QVARS
		R4=A	A		R4[A]=vars
		A=A+A	A		2*vars
		C=A	A
		C=C+C	A		4*vars
		C=C+C	A		8*vars
		C=C+A	A		10*vars
		C=C+CON	A,10		+ DOLIST + SEMI
		GOSBVL	=GETTEMP
		LC(5)	=DOLIST		Start list output
		DAT0=C	A
		AD0EX
		R2=A	A		R2[A] = ->list
		A=A+CON	A,5
		R3=A	A		R3[A] = ->output
		GOSBVL	=D1=DSKTOP
		ST=1	sCNTZERO	No nonzero power yet
		C=0	A
		C=C+1	A		var=1
--		RSTK=C
		A=DAT1	A
		D0=A			D0 = ->P
		A=C	A		A[A] = var
		GOSUBL	CountVarPow	A[A] = power
		?A=0	A
		GOYES	+
		ST=0	sCNTZERO	Flag nonzero power found
+		C=R3	A
		D0=C			D0 = ->output
		LC(5)	=DOBINT
		DAT0=C	A
		D0=D0+	5
		DAT0=A	A
		D0=D0+	5
		CD0EX
		R3=C	A		R3[A] = ->output'
		C=RSTK
		C=C+1	A		var++
		A=R4	A		A[A] = vars
		?C<=A	A
		GOYES	--		Continue until all vars counted for
		C=R3	A
		D0=C
		LC(5)	=SEMI		End list output
		DAT0=C	A		
		A=R2	A		A[A] = ->list
		GOSBVL	=GPPushA
		?ST=0	sCNTZERO	Push TRUE if nonzero power found
		GOYES	+
+		GOVLNG	=PushT/FLoop
ENDCODE

* ::
*    FALSE ZERO 3PICK			( P flag #0 P )
*    Variables				( P F #0 P #v )
*    #1+_ONE_DO (DO)			( P a1 .. ai flag #i P )
* 	INDEX@ VarFactor		( P a1 .. ai flag #i P #n )
* 	4UNROLL SWAP#1+SWAP		( P a1 .. ai ai+1 flag #i+1 P )
* 	4PICK #0=?SKIP
* 	:: ROTDROP TRUE UNROT ;		( P a1 .. ai ai+1 flag' #i+1 P )
*    LOOP
*    DROPSWAP				( P a1 .. an #n flag )
*    OVER#2+UNROLL			( P flag a1 .. an #n )
*    {}N SWAP				( P Lk flag )
* ;

**********************************************************************
* Desc:		Divide polynomial by its trivial powers
* Stack:	( P Lk --> P' )
* Algorithm:	Copy elements of P to output
*		When SEMI is encountered skip back to remove Z0's
*		The number of Z0's == Lk[compositelevel]
**********************************************************************
NULLNAME PDivLk
::
  CODE
		GOSBVL	=SAVPTR

		D1=D1+	5		Allocate duplicate P
		A=DAT1	A
		D0=A			->P
		GOSBVL	=SKIPOB
		CD0EX
		A=DAT1	A
		C=C-A	A
		GOSBVL	=GETTEMP
		AD0EX			->output
		R0=A	A		->output

		GOSBVL	=D1=DSKTOP
		C=DAT1	A
		R1=C	A		->Lk
		D1=D1+	5
		C=DAT1	A		->P
		DAT1=A	A		Overwrite ->P
		D0=C			->input
		D1=A			->output
		B=0	A		complevel

pdivlkloop	A=DAT0	A
		LC(5)	=DOLIST
		?A=C	A
		GOYES	pdivlklist
		LC(2)	=DOHSTR
		?A#C	A
		GOYES	pdivlksemi
		D0=D0+	5
		C=DAT0	A		Zlen
		D0=D0-	5
		C=C+CON	A,5
		GOSBVL	=MOVEDOWN	Copy Z
		GONC	pdivlkloop

pdivlklist	B=B+1	A		complevel++
pdivlkcont	DAT1=C	A		Copy DOLIST
		D0=D0+	5
		D1=D1+	5
		?B#0	A
		GOYES	pdivlkloop
		CD1EX
		D0=C			->Ptail
		A=R0	A		->P
		C=C-A	A		Psize
		C=C+CON	A,5+1		+link+GCnib
		DAT0=C	A
		D0=D0+	5
		GOSBVL	=recover	Remove extra memory
		GOVLNG	=GETPTRLOOP

pdivlksemi	C=R1	A		->Lk
		CD0EX
		RSTK=C			->input saved
		C=B	A
		GOSUB	NthComp#Drop
		GOSUB	PopList#	A[A] = power of this variable
		C=A	A
		C=C+C	A		2*
		A=A+C	A		3*
		C=C+C	A		4*
		C=C+C	A		8*
		A=A+C	A		11*
		CD1EX
		C=C-A	A		Skip back over the Z0's
		CD1EX
		C=RSTK
		D0=C			->input restored
		LC(5)	=SEMI
		B=B-1	A		complevel--
		GONC	pdivlkcont
****************************************
* Find Nth system binary from list
****************************************
NthComp#Drop	D=C	A		N
		D0=D0+	5		->ob1
		D=D-1	A
		RTNC
		LC(5)	=DOBINT
nthcomp#drplp	D=D-1	A
		RTNC
		A=DAT0	A		prolog / ROM address
		D0=D0+	5
		?A#C	A
		GOYES	nthcomp#drplp	Skipped ROM address
		D0=D0+	5
		GONC	nthcomp#drplp	Skipped #
****************************************
* Pop # from list in D0.
****************************************
PopList#	A=DAT0	A
		LC(5)	=SEMI
		?A=C	A
		RTNYES			CS: End of list
		D0=D0+	5
		LC(4)	=DOBINT
		?A=C	A
		GOYES	poplist#
		AD0EX
		D0=D0+	5
		C=DAT0	A		integer
		D0=A
		A=C	A
		RTNCC			CC: Got integer
poplist#	A=DAT0	A		integer
		D0=D0+	5
		RTNCC			CC: Got integer
  ENDCODE
  DROP PTrim
;

**********************************************************************
* Desc:		Calculate what power of the given variable
*		is a factor in P.
* Stack:	( P #var --> P #n )
* Algorithm:
*		When compositelevel == #var
*			then count trailing zeros
*			take minimum with the old result
*		When composite level < #var
*			then any Z <> 0 means power = 0
**********************************************************************
NULLNAME VarFactor
CODE
		GOSBVL	=POP#		A[A] = #var
		GOSBVL	=SAVPTR
		C=DAT1	A
		D0=C			->P
		GOSUB	CountVarPow
		A=A+1	A		Return 0 instead of -1 for
		GOC	+		too high variable number
		A=A-1	A
+		GOVLNG	=PUSH#ALOOP

* Input:	A[A] = #var
*		D0   = ->P
* Output:	A[A] = #pow

CountVarPow	R0=A	A		R0[A] = #var
		A=DAT0	A
		LC(5)	=DOLIST
		?A=C	A
		GOYES	+
		A=0	A		No polynomial - return 0
		RTN
+
		A=0	A		Init minimum to -1
		A=A-1	A
		R1=A	A		R1[A] = min
		B=0	A		complevel

cntpowlist	D0=D0+	5
		B=B+1	A		complevel++
		C=R0	A		#var
		?B=C	A
		GOYES	cntpownow	Count trailing zeros now
cntpowloop	A=DAT0	A
		LC(5)	=DOLIST
		?A=C	A
		GOYES	cntpowlist
		LC(2)	=DOHSTR
		?A#C	A
		GOYES	cntpowsemi
		D0=D0+	5
		C=DAT0	A		Zlen
		AD0EX
		A=A+C	A		Skip Z
		D0=A
		C=C-CON	A,7		CRY set if Z0
		GOC	cntpowloop
		A=0	A		#pow must be zero since Z <> 0
		RTN			(complevel known to be < #var)

cntpowsemi	D0=D0+	5
		B=B-1	A		complevel--
		?B#0	A
		GOYES	cntpowloop	Continue until traversed all
		A=R1	A		min
		A=A+1	A		CRY set if -1
		RTNC			If so then return 0
		A=A-1	A		Else return min
		RTN

cntpownow	D=0	A		zeros = 0
		C=A	W		copy high nibs for Z0 comparisons
		LCZ0
cntpowtrim	A=DAT0	11		First trim the sub poly
		D0=D0+	11		just in case P is untrimmed
		?A=C	W
		GOYES	cntpowtrim
		D0=D0-	11		->first non-zero coeff

cntpownowlp	LC(5)	=DOHSTR		Make sure low 5 nibbles are ok

cntpowzerlp	A=DAT0	11
		?A#C	W
		GOYES	+		Not Z0 - either SEMI or set zeros=0
		D0=D0+	11		Skip Z0
		D=D+1	A		zeros++
		GONC	cntpowzerlp	Count all consecutive zeros
+		LC(5)	=SEMI
		?A=C	A
		GOYES	+		Found SEMI - return zeros counter
		D=0	A		Found nonzero coeff - zeros = 0
		GOSBVL	=SKIPOB		Skip the coeff (possibly not Z!)
		GONC	cntpownowlp	And try zeros again

+		C=R1	A		Check if zeros count is bigger
		?C<=D	A		than current min
		GOYES	cntpowsemi	If so then just continue
		C=D	A
		R1=C	A		Else set new min
		?C#0	A		And continue if var factor is
		GOYES	cntpowsemi	still possible
		A=0	A		Else return 0 as the power
		RTN
ENDCODE
**********************************************************************

