	STITLE	ALG Rational Integration
**********************************************************************
*		ALG Library: Rational function integration
**********************************************************************

**********************************************************************
* EXPORT:
*
*   QInt		( Q --> P Z )
*   #Factorial		( #n --> Z )
*   Horowitz		( q r --> R1 R2 L1 L2 )
*   RatInt  		( sym var -> sym' )
*
**********************************************************************

**********************************************************************
* Name:		QInt
* Desc:		Compute the integral of a polynomial (w.r.t main var)
*		Returns a polynomial and hxs denominator
* Stack:	( Q --> P Z )
**********************************************************************
NULLNAME QInt
::
  DupQIsZero? case HXS 2 10 		( -> 0 1 )
  DUPTYPEHSTR? IT ONE{}N		( A )
  DUPLENCOMP DUP #Factorial		( A n n! )
  1LAMBIND				( A n )
    #1+_ONE_DO (DO)			( A )
       DUPINDEX@ NTHCOMPDROP		( A ai )
       1GETLAM ISTOP-INDEX #>Z QDiv	( A ai ci )
       QMul SWAP			( bi A  )
    LOOP				( b1..bn A )
    LENCOMP 				( b1..bn n )
    HXS 1 0 SWAP#1+ {}N			( B )
  1GETABND				( B n! )
  RCanno				( P Z )
;

**********************************************************************
* Name:		#Factorial
* Desc:		Compute n factorial
* Stack:	( #n --> Z )
**********************************************************************
NULLNAME #Factorial
::
  HXS 2 10 SWAP			( 1 #n )
  DUP#0=csDROP			( -> 1 )
  #1+_ONE_DO (DO)		( Z )
    INDEX@ #>Z			( Z i )
    QMul			( Z' )
  LOOP
;
    
**********************************************************************
* Name:		Horowitz
* Desc:		Horowitz reduction for rational integral
*		   Int(P/Q) -> R1/R2 + Int(L1/L2) 
*		   deg(L1)<deg(L2) & L2 is square-free
* Stack:	( P Q --> R1 R2 L1 L2 )
**********************************************************************
NULLNAME Horowitz
::

  ( * compute the polynomial part of the integral * )

  EuclidDiv				( a b q r )
  4ROLL QInt 				( b q r P1 Z )
  5ROLL QMul 				( q r P1 P2 )
  2SWAP 				( P1 P2 q r )
  
  ( * done if no remainder * )
 
  OVER QIsZero? ?SEMI			( -> P1 P2 0 1 )

  ( * o/w horowitz algo * )

  DUPDUP PDer QGcd			( P1 P2 q r r1 )

  ( * quick exit if r is square-free * ) 

  DUPTYPELIST? ITE DUPLENCOMP ONE	( P1 P2 q r r1 1+deg )
  #1= caseDROP 				( -> P1 P2 q r )

  ( * calculate other variables * )

  2DUP QDiv				( P1 P2 q r r1 r2 )
  DUP3PICK PDer QMul 3PICK QDiv		( P1 P2 q r r1 r2 s )
  DUP DUPTYPELIST? ITE INNERCOMP ONE	( P1 P2 q r r1 r2 s s1..sn n )
  HXS 1 0 SWAP#1+ {}N			( P1 P2 q r r1 r2 s sx )
  4PICK LENCOMP #1-			( P1 P2 q r r1 r2 s sx n1 )
  4PICK LENCOMP #1-			( P1 P2 q r r1 r2 s sx n1 n2 )

  {{ p1 p2 q r r1 r2 s sx n1 n2 }}	( )

  ( * create empty linear system * )

  n1 n2 #+ DUP#1+ MMnull		( MM )

  ( Put q to last column )

  q ZERO 4PICK #1- MMputcol		

  ( * put q1 columns for first n1 columns * )

  ::
    n1 #1- DUP#0=csDROP
    ZERO_DO (DO)
      ISTOP-INDEX #>Z	 		( MM y )
      r2 QMul sx QSub			( MM y*r2-s*x )
      ISTOP-INDEX #1-			( MM y*r2-s*x #y )
      INDEX@ MMputcol			( MM' )	
    LOOP			
  ;

  s QNeg ZERO n1 #1- MMputcol		

  ( * put q2 columns for the next n2 columns * )

  n2 ZERO_DO (DO)
    r1 INDEX@ 4PICK OVER#- #2-
    MMputcol
  LOOP

  ( * solve the system * )

  MMsolve

  ( * extract q1 * )

  n1 ZERO MMgetR			( MM q1 d1 )
  r1 QMul				( MM q1 d1*r1 )
  p1 p2 RAdd RCanno			( MM R1 R2 )

  ( * extract q2 * )

  5PICK n1 MMgetR			( MM R1 R2 q2 d2 )
  r2 QMul RCanno			( MM R1 R2 L1 L2 )

  ( * done * )

  MMdrop				( R1 R2 L1 L2 )
  ABND
;  

**********************************************************************
* Name:		RatInt
* Desc:		Integral of a rational symbolic (w.r.t main var)
* Stack:	( sym var -> sym' )
**********************************************************************

* From algrsymb.s
*
* DEFINE Rvars@	3GETLAM
* DEFINE Rmain@	2GETLAM

NULLNAME RatInt
::
  RSimVars 				( P Q v )

  ( * check for 'safe' integration * )

  DUPDUP CARCOMP			( P Q v v X )
  RESOROMP NotSafeInt? POSCOMP		( P Q v #k )
  #0= NcaseSIZEERR			( P Q v )

  ( * split integral of P/Q using Horowitz method * )

  UNROT Horowitz			( v R1 R2 L1 L2 )
  2SWAP 5PICK R2A			( v L1 L2 ar )
  4UNROLL PFDecomp			( ar v Lpf )

  ( * done if no logaritmic part * )

  DUPNULL{}? case2DROP			( -> ar )

  ( * process factors of logarithmic part * )

  ZEROSWAP				( ar v #main Lpf )
  { NULLLAM NULLLAM NULLLAM }
  BIND					( ar )
  1GETLAM LENCOMP 			
  #1+_ONE_DO (DO)			
    ::					( sym )
      1GETLAM INDEX@ NTHCOMPDROP	( sym pfi )
      INCOMPDROP DROP			( sym nom den q )
      DUPLENCOMP #3= case RQuadInt	( * quadratic term *)
      
      3PICKOVER PDer			( sym nom den q nom q' )
      PseudoDiv SWAP			( sym nom den q s c r )
      QIsZero? NOTcase2drop RNoInt	( * not a simple log form *)

      ( * term is a simple logarithm * )

      5ROLLDROP 4ROLL			( sym q s c den )
      QMul RCanno			( sym q fnom fden )
      RGetSign R2SF			( sym q sflag Mf )
      DUP #3+ ROLL			( sym sflag Mf q )
      HXS 2 10 QZ2S			( sym sflag Mf Mq )
      putop: xLN			( sym sflag Mf Mlnq )
      SMul CKSymbN			( sym sflag aflnq )
      SWAP ITE XSub XAdd		( sym' )
    ;
  LOOP
  ABND
;

*---------------------------------------------------------------------
* Name:		NotSafeInt?
* Desc:		TRUE if obj is not safe to integrate w.r.t. X
* Note:		
*		obj is consider not safe if either:
*		
*		a) obj is a symbolic containing X
*		b) obj is 'i', the complex unit
*		c) obj is the square-root of a number
*
* Stack:	( obj X -> flag )
*---------------------------------------------------------------------
NULLNAME NotSafeInt?
::
  OVER TYPESYMB? NOTcase2drop FALSE	( -> FALSE is obj not symb )
  Contains? SWAPDROP casedrptru		( -> TRUE if case a )
  					( symb )
  DUP DUPLENCOMP NTHCOMPDROP		( symb ob )
  DUP EQ: xi case2drop TRUE		( -> TRUE if case b )
  EQ: xSQRT NOTcasedrop FALSE		( -> FALSE if not SQRT )
  DUPLENCOMP #2<> casedrpfls		( -> FALSE if not SQRT[simple] )
  CARCOMP TYPEREAL?			( -> flag )
;

*---------------------------------------------------------------------
* Name:		RNoInt
* Desc:		Add term to sym for non-integration of frac num/den*q 
* Stack:	( sym num den q --> sym' )
* Note:		Assumes LAM vars is bound
*---------------------------------------------------------------------
NULLNAME RNoInt		
::
  ONE Frac2SS				( sym sflag Mterm )
  Rvars@ CARCOMP syminner		( sym sflag Mterm Mvar )
  SYMBOL ID INT ; TWO 
  ' xFCNAPPLY THREE 			( sym sflag Mterm Mvar Mint )
  top& top& CKSymbN			( sym sflag aint )
  SWAP ITE XSub XAdd			( sym' )
;

*---------------------------------------------------------------------
* Name:		RQuadInt
* Desc:		Add term to sym for integration of fraction
*		num/den*q with q quadratic
* Stack:	( sym num d q --> sym' )
*
* Note:		LAM vars and Deno are bound
*		q is primitive, positive, irreducible
*		deg(num)<2 
*---------------------------------------------------------------------

NULLNAME RQuadInt	
::
  DUP INCOMPDROP 6ROLL 			( sym d q a b c num )
  DUPTYPELIST? ITE INNERCOMP ONE	( sym d q a b c [e f 2/f 1] )
  #1= ITE HXS 1 0 SWAP			( sym d q a b c f e )
  HXS 1 0 DUP Rvars@ Rmain@		( sym d q a b c f e u0 v0 vars #main )
  ZERO					( * Dummy object * )
  {{ d q a b c f e u v vars main tmp }}	( sym )

  ( * add ln[q] term * )
  
  e QIsZero? ?SKIP RQILn		( sym' )

  ( * compute coefficient numerator s1 * )

  a f QMul HXS 2 20 QMul		( sym 2af )
  e b QMul QSub PTrim			( sym 2af-eb )
  ONE CKMkP				( sym s1 )

  ( * done if zero * )
 
  DupQIsZero? casedrop ABND		( -> sym ) 

  ( * compute discriminant w=b^2-4*a*c * )
  
  b DUP QMul a c QMul			( sym s1 b^2 ac )
  HXS 2 40 QMul QSub PTrim		( sym s1 w )
  DUP PLCZ ZIsNeg?			( sym s1 w sf )
  DUP4UNROLL IT QNeg			( sym sf s1 |w| )

  ( * compute coefficient denominator s2 * )

  TWO QRoot 2DUP ONE CKMkP	 	( sym sf sq u v u v' )
  v! u!					( sym sf s1 u v )
  QMul a QMul ONE CKMkP			( sym sf s1 uva )
  d QMul				( sym sf s1 s2 )

  ( * remove common factor from 2a b u * )

  a HXS 2 20 QMul 			( sym sf s1 s2 2a )
  b u THREE{}N PPP DROP			( sym sf s1 s2 { 2a' b' u' } )
  INCOMPDROP ONE CKMkP			( sym sf s1 s2 2a' b' u'' )
  u! b! a!				( sym sf s1 s2 )

  ( * Note:  s1, s2, u, v - has var x1, x2 ... * )
  ( *        2a, b,       - has var x2, ...    * )

  ( * add atan or log term depending on sign of discriminant * )

  ROT ITE RQINeg RQIPos 		( sym' )

  ( * done * )

  ABND
;

* Add logarithmic term  e/2ad*log[q] 

NULLNAME RQILn		( sym --> sym' )
::
  e ONE CKMkP				( e )
  a HXS 2 20 QMul 		 	( e 2a )
  ONE CKMkP d QMul 			( e 2ad )
  RCanno RGetSign R2SF			( sym sflag Mf )
  q HXS 2 10 PZ2S			( sym sflag Mf Mq )
  putop: xLN				( sym sflag Mf Mlnq )
  SMul CKSymbN				( sym sflag aflnq )
  SWAP ITE XSub XAdd			( sym' )
;

* Add two log's for integration of inv[q] with b^2-4ac positive

NULLNAME RQIPos		( sym s1 s2 --> sym' )
::
  HXS 2 20 QMul RCanno			( sym s1 2s2 )
  RGetSign R2SF				( sym sflag Ms )
  RQIxSqrtv CKSymbN			( sym sflag af )
  2DUP f! e!				( sym sflag af )

  ( log[2ax+b-u*sqrt[v]] term )

  a b TWO{}N				( sym sflag af 2ax+b )
  HXS 2 10 PZ2S				( sym sflag af Mq' )
  u HXS 2 10 R2SF			( sym sflag af Mq' Mu )
  RQIxSqrtv SSub			( sym sflag af Mq'-usqrtv )
  putop: xLN				( sym sflag af Mln )
  CKSymbN XMul				( sym sflag afln )
  SWAP ITE XSub XAdd			( sym' )

  ( log[2ax+b+u*sqrt[v]] term )

  e f	 				( sym sflag af )
  a b TWO{}N				( sym sflag af 2ax+b )
  HXS 2 10 PZ2S				( sym sflag af Mq' )
  u HXS 2 10 R2SF			( sym sflag af Mq' Mu )
  RQIxSqrtv SAdd			( sym sflag af Mq'+usqrtv )
  putop: xLN				( sym sflag af Mln )
  CKSymbN XMul				( sym sflag afln )
  SWAP ITE XAdd	XSub			( sym' )
;

* Add atan term for integration of inv[q] with b^2-4ac negative

NULLNAME RQINeg		( sym s1 s2 --> sym' )
::
  RCanno				( sym s1' s2' )
  RGetSign R2SF				( sym sflag Ms )
  RQIxSqrtv 				( sym sflag Mf )
  a b TWO{}N				( sym sflag Mf 2ax+b )
  HXS 2 10 PZ2S				( sym sflag Mf Mq' )
  u HXS 2 10 R2SF			( sym sflag Mf Mq' Mu )
  RQIxSqrtv				( sym sflag Mf Mq' Muv )
  DUP#0=ITE DROP SDiv			( sym sflag Mf Min )
  putop: xATAN				( sym sflag Mf Matan )
  SMul CKSymbN				( sym sflag afatan )
  SWAP ITE XSub XAdd			( sym' )
;

* Multiply Meta by square-root of v

NULLNAME RQIxSqrtv	( Meta -> Meta' )
::
  v DupZIsOne? caseDROP			( -> Meta )
  HXS 2 10 QZ2S				( Meta Mv )
  putop: xSQRT				( Meta Msqrtv )
  SMul					( Meta' )
;

**********************************************************************

