	STITLE	ALG Assembly Utilities

**********************************************************************
*	ALG Library:	Small Assembly Utilities
**********************************************************************

**********************************************************************
* EXPORT:
*	SHORT-		Short signed integer substraction
*	SHORT+		Short signed integer addition
*	SHORT*		Short signed integer multiplication
*	SHORT/		Short signed integer division
*	BDIV		Short unsigned integer division
*	SHORTGCD	Short unsigned integer gcd
*	SHORTCANNO	Short unsigned rational simplification
*	SHORTMULMOD	Short unsigned modular multiplication
*	SHORTPOWMOD	Short unsigned modular power
*	SHORTOUT	Short signed integer output
*	SHORTPOP	Short signed integer pop
*	SHORTPOPRAT?	Attempt short signed rational pop from stack
*	SHORTPOPZ?	Attempt short signed integer pop from stack
*	SHORTPOP?	Attempt short signed integer pop from D0
*	SHORTPOPD1	Attempt short signed integer pop from D1
*
*	QLEN		Polynomial number of elements calculation
*	QLEN2		Same for 2 polynomials
*	QSIZE		Polynomial size calculation
*
*	QCOPY		Polynomial copy to output
*	QCOPYREM	Polynomial copy to output (remainder of Q)
*
*	QSKIP		Skip polynomial
*	QSKIPREM	Skip remainder of polynomial
*	QSKIPD1		Skip polynomial in D1
*
*	QVARS		Count variables in a polynomial
*	QVARS2		Count variables in two polynomials
*
*	EQUAL?		General object equality test
*	CONTAINS?	General embedded object test
*
*	PushN3Loop	3DROP, Push N, Loop
*	PushZ3Loop	3DROP, Push Z, Loop
*	PushN2Loop	2DROP, Push N, Loop
*	PushZ2Loop	2DROP, Push Z, Loop
*	PushN1Loop	DROP, Push N, Loop
*	PushZ1Loop	DROP, Push Z, Loop
*	PushNLoop	Push N, Loop
*	PushZLoop	Push Z, Loop
*	PushN		Push N
*	PushZ		Push Z
*	PushZ0		Push Z0
*	PushZ1		Push Z1
*
*	UNIVAR?		Test if polynomial is univariate
*	SUNIVAR?	Test if polynomial is short and univariate
*
*	QSHORT?		Test if polynomial has short coefficients
*	QSHORT2?	Test if stack contains two short polynomials
*
*	ShrinkQ		Shrink the allocated memory to contain only Q
*	TrashTop	Trash topmost object in tempob area
*
* The idea is to centralize the locations of essential general purpose
* subroutines so that close-by modules can all use them.
*
**********************************************************************

ASSEMBLE

**********************************************************************
* Desc:		Substract/add small integers with overflow checks
* Entry:	A[W] B[S] = signed integer X
*		C[W] D[S] = signed integer Y
* Output:	A[W] C[S] = signed integer X+Y or X-Y
*		CS if overflow, CC otherwise
* Modifies:	A[W] C[S] D[S]
* Stack:	0
**********************************************************************

SHORT-		D=-D-1	S	Negate Y

SHORT+		CDEX	S
		?C#B	S
		GOYES	+	Handle different signs
		CDEX	S
		A=A+C	W	X+Y
		RTNC		CS if overflow
-		C=B	S	Return X sign as X+Y sign
		RTNCC
+		CDEX	S
		A=A-C	W
		GONC	-	Substracted correctly - done
		A=-A	W	Fix to absolute value
		C=D	S	Return Y sign
		RTNCC

**********************************************************************
* Desc:		Multiply small integers with overflow checks
* Entry:	A[W] B[S] = signed integer X
*		C[W] D[S] = signed integer Y
* Output:	A[W] C[S] = signed integer X*Y
* Modifies:	A[W] B[W] C[W] D[S]
* Stack:	0
**********************************************************************

SHORT*		CDEX	S		Calculate result sign
		C=C+B	S
		GONC	+
		C=0	S
+		CDEX	S		D[S] = result sign
		B=0	W		result = 0
short*loop	SB=0
		CSRB			y /= 2
		?SB=0
		GOYES	+
		B=B+A	W		result += x
		RTNC			CS: Overflow
+		A=A+A	W		x *= 2
		RTNC			CS: Overflow
		?C#0	B		\
		GOYES	short*loop	 Faster than using only the [W] test
		?C#0	W
		GOYES	short*loop	/
		A=B	W		A[W]=|X*Y|
		C=D	S		C[S]=sgn(X*Y)
		RTNCC			CC: Valid result

**********************************************************************
* Desc:		Multiply small integers with no overflow checks
* Entry:	A[W] = multiplicand
*		C[W] = multiplier
* Exit:		A[W] = C[W] = result
* Modifies:	A[W] B[W] C[W]
* Stack:	0
**********************************************************************

BMUL		B=C	W
BMULB		C=0	W
bmulnow		?A<B	W		Get smaller to A[W] for speed
		GOYES	+
		ABEX	W
+		SB=0
bmulloop	ASRB
		?SB=0
		GOYES	+
		SB=0
		C=C+B	W
+		B=B+B	W
		?A#0	B
		GOYES	bmulloop
		?A#0	W
		GOYES	bmulloop
		A=C	W
		RTNCC

**********************************************************************
* Desc:		Multiply small integers with carry
* Entry:	A[W]    = multiplicand
*		B[W]    = multiplier
*		C[8-15] = carry
* Exit:		A[W] = C[W] = result
* Modifies:	A[W] B[W] C[W]
* Stack:	0
**********************************************************************

BMULCRY		CSR	W
		CSR	W
		CSR	W
		CSR	W
		CSR	W
		CSR	W
		CSR	W
		CSR	W		C[0-7]=carry
		GOTO	bmulnow		Jump to common code

**********************************************************************
* Desc:		Divide small integers with overflow checks (never!)
* Entry:	A[W] B[S] = signed dividend
*		C[W] D[S] = signed divisor
* Output:	A[W] = quotient
*		B[W] = remainder
*		C[W] = divisor
*		D[S] = quotient sign	!!! differs from the rest !!!
*		CS if division by zero
* Modifies:	A[W] B[W] D[S]
* Stack:	1 (D[A] save)
**********************************************************************

SHORT/		CDEX	S		Calculate quotient sign
		C=C+B	S
		GONC	+
		C=0	S
+		CDEX	S		D[S] = quotient sign

* Special unsigned division entry.

BDIV		B=A	W		B[W] = dividend
		A=C	W		C[W] = divisor
		?C=0	W
		RTNYES			CS: Division by zero
		CDEX	A		Save D[A]
		RSTK=C
		CDEX	A
		D=0	B		D[B]=#bits of quotient
bdivloop1	?A>B	W		Align divisor
		GOYES	+
		D=D+1	B		bits++
		C=A	W
		A=A+A	W
		GONC	bdivloop1
+		A=0	W		A[W]=quotient
		D=D-1	B
		GOC	bdivdone
bdivloop2	?C>B	W
		GOYES	+
		B=B-C	W
		A=A+1	B
+		D=D-1	B
		GOC	bdivdone
		CSRB
		A=A+A	W
		GONC	bdivloop2
bdivdone	CDEX	A		Restore D[A]
		C=RSTK
		CDEX	A
		RTNCC			CC: Not division by zero

**********************************************************************
* Desc:		Integer GCD using the binary algorithm
* Entry:	A[W] = X
*		C[W] = Y
* Exit:		A[W] = GCD(X,Y)
* Modifies:	A[W] C[W] D[A]
* Stack:	0
* Algorithm:	gcd(2n,2m)	= 2gcd(n,m)
*		gcd(2n,2m+1)	= gcd(n,2m+1)
*		gcd(2n+1,2m+1)	= gcd(max-min,min)
*
**********************************************************************

SHORTGCD	D=0	A		pow2=0

		?A=0	W		Handle gcd(0,x)=x
		GOYES	gcd64_0
		ACEX	W
		?A=0	W
		GOYES	gcd64_0

-		?ABIT=1	0		Handle gcd(2n,2m)=2gcd(n,m)
		GOYES	gcd64_odd
		ACEX	W
		?ABIT=1	0
		GOYES	gcd64_odd
		ASRB
		CSRB
		D=D+1	A		pow2++
		GONC	-

gcd64_odd	?CBIT=1	0		Handle gcd(2n,2m+1)=gcd(n,2m+1)
		GOYES	gcd64_max
		CSRB
		GONC	gcd64_odd

gcd64_max	?C<A	W		Handle gcd(2n+1,2m+1)=gcd(max-min,min)
		GOYES	+
		ACEX	W
+		A=A-C	W		max-min
		?A=0	P
		GOYES	+		try exit if A[0]=0

-		ASRB			odd-odd is even
		?ABIT=0	0
		GOYES	-
		GONC	gcd64_max

+		?A#0	W		exit fails?
		GOYES	-		yes - then continue

gcd64_0		A=C	W		gcd
		?D=0	A
		RTNYES			pow2=0 - done
-		A=A+A	W
		D=D-1	A
		?D#0	A
		GOYES	-		double until pow2 = 0
		RTN

**********************************************************************
* Desc:		Simplify short rational
* Entry:	A[W]=n1		C[W]=n2
* Exit:		A[W]=n1'	C[W]=n2'
* Stack:	2
* Alters:	A[W] B[W] C[W] D[W]
* Note:		The 2's are divided out directly since calling
*		SHORTGCD would require using an extra register
*		which the MS subroutines cannot afford.
**********************************************************************

SHORTCANNO	?C=0	W		x/0 = x/0
		RTNYES
		?A#0	W
		GOYES	zcannolp1
		C=0	W		0/x = 0/1
		C=C+1	A
		RTN
* First divide out 2's
zcannolp1	?ABIT=1	0
		GOYES	zcannock1
		?CBIT=1	0
		GOYES	zcannock1
		ASRB
		CSRB
		GONC	zcannolp1
* Now check 1/X and X/1 for speed
zcannock1	A=A-1	W
		?A#0	W
		GOYES	+
		A=A+1	W		1/x = 1/x
		RTN
+		A=A+1	W
		C=C-1	W
		?C#0	W
		GOYES	+
		C=C+1	W		x/1 = x/1
		RTN
+		C=C+1	W
* Now calculate the GCD of the integers
		B=A	W		B[W] = x
		D=C	W		D[W] = y
		GONC	+		Jump into the loop
zcannolp2	CSRB			Divide 2's out from C[W]
+		?CBIT=0	0
		GOYES	zcannolp2
		GONC	+		Jump into the next loop
zcannolp3	ASRB			Divide 2's out from A[W]
+		?ABIT=0	0
		GOYES	zcannolp3
		?A>=C	W		Order max(x,y) to A[W]
		GOYES	+
		ACEX	W
+		A=A-C	W		A[W]=max-min
		?A#0	W
		GOYES	zcannolp3	Divide 2's out from A[W] again
* Now check the GCD for trivial 1
		C=C-1	W
		?C#0	W
		GOYES	+
		A=B	W		Cannot simplify
		C=D	W
		RTN
* Divide GCD from both numbers
+		C=C+1	W
		A=B	W
		GOSUB	BDIV		A[W]= x/gcd
		CDEX	W
		ACEX	W
		CDEX	W
		GOSUB	BDIV		A[W] = y/gcd
		C=D	W
		ACEX	W
		RTN

**********************************************************************
* Desc:		Short unsigned integer modular multiplication
* Entry:	A[W] = X
*		C[W] = Y
*		B[W] = N
* Exit:		A[W]=C[W]= Z = X*Y MOD N
* Modifies:	A[W] C[W] D[W]
* Stack:	0
**********************************************************************

SHORTMULMOD	D=C	W		D[W] = Y
		C=0	W		Z = 0
		SB=0
smulmodloop	DSRB			Y>>1
		?SB=0
		GOYES	smulmodnext	No addition of term
		SB=0
		C=C+A	W		Z += X
		GOC	+		Handle overflow
		?C<B	W
		GOYES	smulmodnext
+		C=C-B	W		Z -= N
smulmodnext	A=A+A	W		X << 1
		GOC	+		Handle overflow
		?A<B	W
		GOYES	smulmodcont
+		A=A-B	W		X -= N
smulmodcont	?D#0	B		\
		GOYES	smulmodloop	 Faster than plain [W] test
		?D#0	W
		GOYES	smulmodloop	/
		A=C	W
		RTNCC

**********************************************************************
* Desc:		Short unsigned integer modular power
* Entry:	A[W] = X
*		C[W] = Y
*		B[W] = N
* Exit:		A[W] = Z = X^Y MOD N
* Modifies:	A[W] C[W] D[W] R0[W] R1[W]
* Stack:	1 (SHORTMULMOD call)
**********************************************************************

SHORTPOWMOD	R1=A	W		R1[W] = X
		A=0	W
		A=A+1	P		Z = 1
spowmodloop	SB=0
		CSRB			Y >> 1
		R0=C	W		R0[W] = Y
		?SB=0
		GOYES	+
		C=R1	W		C[W] = X
		GOSUB	SHORTMULMOD	A[W] = Z*X MOD N = Z
+		AR1EX			A[W]=X	R1[W]=Z
		C=A	W		C[W]=X
		GOSUB	SHORTMULMOD	A[W] = X^2 MOD N = X
		AR1EX			A[W]=Z	R1[W]=X
		C=R0	W		C[W]=Y
		?C#0	W
		GOYES	spowmodloop
		RTNCC

**********************************************************************
* Desc:		Output signed small integer (trimmed)
* Entry:	A[W] C[S] = signed integer
*		D1 = ->output
* Exit:		A[W] C[S] unchanged
*		D1 = ->output tail
*		CRY clear
* Modifies:	C[1..6] D1 P=0
* Stack:	0
**********************************************************************

SHORTOUT	P=	0		Safety opcode
		LC(5)	=DOHSTR		Output prolog
		DAT1=C	A
		D1=D1+	5
		?A=0	W
		GOYES	shortout0

		P=	15		Trimming loop
-		?A#0	P
		GOYES	+
		P=P-1
		GONC	-
+		C=0	A		A is nonzero
		C=P	0
		C=C+CON	B,7
		DAT1=C	A		output length
		D1=D1+	5
		DAT1=A	WP		output mantissa
		CD1EX
		C+P+1
		CD1EX
		P=	0
		DAT1=C	S		output sign
		D1=D1+	1
		RTNCC

shortout0	LC(6)	6
		DAT1=C	6
		D1=D1+	6
		RTNCC

**********************************************************************
* Desc:		Pop signed small integer (no size checks!)
* Entry:	D0 = ->integer
* Exit:		A[W] C[S] = signed integer
*		D0 = ->integer tail
*		CS: integer is zero
*		CC: integer is nonzero
* Modifies:	A[W] C[A] D0
* Stack:	0
**********************************************************************

SHORTPOP	D0=D0+	5		Skip DOHSTR
shortpop	C=DAT0	A		Zlength
		D0=D0+	5
		A=0	W		Initialize high nibbles
		C=C-CON	B,7
		GONC	shortpop#
		D0=D0+	1
		RTNSC			CS: Z is zero

shortpop#	P=C	0
		A=DAT0	WP		Pop mantissa
		CD0EX
		C+P+1			Skip mantissa
		CD0EX
		P=	0
		C=DAT0	S		Pop sign
shortpop=	D0=D0+	1
		RTNCC			CC: Z is nonzero

**********************************************************************
* Desc:		Pop signed small integer with size check from stack
* Entry:	D1 = ->stack level
* Exit:		CS: Too big integer
*		CC: Integer fits into register
*			D0 = ->Ztail
*			A[W] C[S] = Z2
*			B[W] D[S] = Z1
* Modifies:	A[W] B[W] C[W] D[S] D0 D1
* Stack:	1
**********************************************************************
SHORTPOPRAT?	GOSUB	SHORTPOPZ?
		RTNC			CS: Fail
		B=A	W
		D=C	S

**********************************************************************
* Desc:		Pop signed small integer with size check from stack
* Entry:	D1 = ->stack level
* Exit:		CS: Too big integer
*		CC: Integer fits into register
*			D0 = ->Ztail
*			A[W] C[S] = Z
* Modifies:	A[W] C[A] D0 D1
* Stack:	0
**********************************************************************
SHORTPOPZ?	A=DAT1	A
		D1=D1+	5
		D0=A
SHORTPOPCK?	A=DAT0	A
		LC(5)	=DOHSTR
		?A#C	A
		RTNYES
**********************************************************************
* Desc:		Pop signed small integer with size check
* Entry:	D0 = ->integer
* Exit:		CS: Too big integer
*		CC: Integer fits into register
*			D0 = ->Ztail
*			A[W] C[S] = Z
* Modifies:	A[W] C[A] D0
* Stack:	0
**********************************************************************
SHORTPOP?	D0=D0+	5		Skip DOHSTR
shortpop?	A=DAT0	A		Zlength
		D0=D0+	5
		LC(5)	5+16+1		prolog + word + sign
		?A>C	A
		RTNYES			CS: Too long integer
		C=A	A		Zlength
		A=0	W		Init high nibbles
		C=C-CON	B,7
		GONC	shortpop#	Nonzero - continue at common code
		GOC	shortpop=	Zero - skip 1 nibble

**********************************************************************
* Desc:		Pop signed small integer (no size checks!)
* Entry:	D1 = ->integer
* Exit:		A[W] C[S] = signed integer
*		D1 = ->integer tail
*		CS: integer is zero
*		CC: integer is nonzero
* Modifies:	A[W] C[A] D1
* Stack:	0
**********************************************************************

SHORTPOPD1	D1=D1+	5		Skip DOHSTR
shortpopd1	A=0	W		Initialize high nibbles
		C=DAT1	A		Zlength
		D1=D1+	5
		C=C-CON	A,7
		GONC	shortpopd1#
		D1=D1+	1
		RTNSC			CS: Z is zero

shortpopd1#	P=C	0
		A=DAT1	WP		Pop mantissa
		CD1EX
		C+P+1			Skip mantissa
		CD1EX
		P=	0
		C=DAT1	S		Pop sign
		D1=D1+	1
		RTNCC			CC: Z is nonzero

**********************************************************************
* Desc:		Calculate elements in two polynomials
* Entry:	D0 = ->list1	A[A] = len1
*		D1 = ->list2	C[A] = len2
* Exit:		D0 = ->list1
*		D1 = ->list2
* Modifies:	A[A] B[A] C[W] D[A]
* Stack:	2 (PLEN call + save)
**********************************************************************

QLEN2		CD1EX			Swap list1 and list2
		CD0EX
		CD1EX
		GOSUB	QLEN		C[A] = len2
		RSTK=C			RSTK = len2
		CD1EX			Swap list1 and list2 back
		CD0EX
		CD1EX
		GOSUB	QLEN		C[A] = len1
		A=C	A		A[A] = len1
		C=RSTK			C[A] = len2
		RTNCC

**********************************************************************
* Desc:		Calculate number elements in polynomial
* Entry:	D0 = ->P
* Exit:		D0 = ->P
*		A[A]=B[A]=C[A] = length
* Modifies:	A[A] B[A] C[W] D[A]
* Stack:	0
**********************************************************************

QLEN		B=0	A		N=0
		CD0EX
		D0=C
		CSL	W
		CSL	W
		CSL	W
		CSL	W
		CSL	W		C[A1] = ->P
		D0=D0+	5		Skip DOLIST
		D=0	A		complevel = 0
plenloop	A=DAT0	A
		D0=D0+	5
		LC(5)	=SEMI
		?A=C	A
		GOYES	plensemi
		?D#0	A		elements++ if not embedded
		GOYES	+
		B=B+1	A
+		LC(4)	=DOLIST		Skip done if list
		?A=C	A
		GOYES	plenlist
		C=DAT0	A
		AD0EX
		A=A+C	A		Skip Z
		AD0EX
		GONC	plenloop
plenlist	D=D+1	A		complevel++
		GONC	plenloop
plensemi	D=D-1	A		complevel--
		GONC	plenloop
		CSR	W
		CSR	W
		CSR	W
		CSR	W
		CSR	W
		D0=C			->P
		A=B	A		elements
		C=A	A		elements
		RTNCC

**********************************************************************
* Desc:		Return size of Q
* Entry:	D0 ->object
* Exit:		D0 ->object
*		A[A] ->object tail
*		C[A] obsize
* Modifies:	A[W] C[A]
* Stack:	2 (QSKIP call+save)
**********************************************************************

QSIZE		CD0EX
		RSTK=C			C[A] = ->Q
		D0=C
		GOSUB	QSKIP		D0 = ->Qtail
		C=RSTK			C[A] = ->Q
		AD0EX			A[A] = ->Qtail
		D0=C
		C=A-C	A		C[A] = Qsize
		RTNCC

**********************************************************************
* Desc:		Copy Q to output
* Entry:	D0 = ->Q
*		D1 = ->output
* Exit:		D0 = ->Qtail
*		D1 = ->output tail
* Modifies:	A[W] C[A] D0 D1
* Stack:	2 (QSKIP call + save)
**********************************************************************

QCOPY		CD0EX
		RSTK=C
		D0=C
		GOSUB	QSKIP		D0 = ->Qtail
qcopycont	C=RSTK
		AD0EX			->Qtail
		D0=C			->Q
		C=A-C	A		Qsize
		GOVLNG	=MOVEDOWN

**********************************************************************
* Desc:		Copy remainder of Q to output
* Entry:	D0 = ->element in Q
*		D1 = ->output
* Exit:		D0 = ->Qtail
*		D1 = ->output tail
* Modifies:	A[W] C[W] D0 D1
* Stack:	2 (QSKIPREM call + save)
**********************************************************************

QCOPYREM	CD0EX
		RSTK=C
		D0=C
		GOSUB	QSKIPREM	D0 = ->Qtail
		GOTO	qcopycont	Same code as for QCOPY continues

**********************************************************************
* Desc:		Skip to Nth element in Q
* Entry:	B[A] = N-1
*		D0   = ->Q
* Exit:		D0   = ->Q[N]
*		C[A] = ->Q
* Modifies:	A[A] B[A] C[W] D[A]
* Stack:	0
**********************************************************************

QSKIPN		CD0EX
		D0=C
		CSL	W		Save D0 in C[A1]
		CSL	W
		CSL	W
		CSL	W
		CSL	W
		D0=D0+	5		Skip DOLIST
		D=0	A		complevel=0
qskipnloop	?D#0	A
		GOYES	qskipn+
		B=B-1	A		N--
		GOC	qskipnok	Found Nth element - finish up
qskipn+		A=DAT0	A
		D0=D0+	5
		LC(5)	=SEMI
		?A=C	A
		GOYES	qskipnsemi
		LC(4)	=DOLIST
		?A=C	A
		GOYES	qskipnlist
		C=DAT0	A		Zlen
		AD0EX
		A=A+C	A
		AD0EX
		GONC	qskipnloop
qskipnlist	D=D+1	A		complevel++
		GONC	qskipn+
qskipnsemi	D=D-1	A
		GONC	qskipnloop
		RTNSC			CS: Too big N

qskipnok	CSR	W		Restore D0 from C[A1]
		CSR	W
		CSR	W
		CSR	W
		CSR	W
		RTNCC			CC: N was ok

**********************************************************************
* Desc:		Skip remainder of Q in D0
* Entry:	D0 = ->element in Q
*		D0 = ->Qtail
* Modifies:	A[A] C[A] D0
* Stack:	0
**********************************************************************

QSKIPREM	P=	10-1		Save D[A] in D[A1]
		DSL	WP
		DSL	WP
		DSL	WP
		DSL	WP
		DSL	WP
		D=D+1	A		complevel=1
		GONC	+

**********************************************************************
* Desc:		Skip Q in D0
* Entry:	D0 = ->Q
* Exit:		D0 = ->Qtail
* Modifies:	A[A] C[A] D[A1] D0
* Stack:	0
**********************************************************************

QSKIP		P=	10-1		Save D[A] in D[A1]
		DSL	WP
		DSL	WP
		DSL	WP
		DSL	WP
		DSL	WP		D[A]=complevel=0
+		P=	0
qskiploop	A=DAT0	A
		D0=D0+	5
		LC(5)	=DOHSTR
		?A=C	A
		GOYES	qskiphstr
		LC(2)	=DOLIST
		?A#C	A
		GOYES	qskipsemi
		D=D+1	A		complevel++
		GONC	qskiploop
qskiphstr	A=DAT0	A
		CD0EX
		C=C+A	A
		CD0EX
		?D#0	A
		GOYES	qskiploop
qskipend	P=	10-1		Restore D[A] from D[A1]
		DSR	WP
		DSR	WP
		DSR	WP
		DSR	WP
		DSR	WP
		P=	0
		RTNCC
qskipsemi	D=D-1	A
		?D#0	A
		GOYES	qskiploop
		GONC	qskipend

**********************************************************************
* Desc:		Skip Q in D1
* Entry:	D1 = ->Q
* Exit:		D1 = ->Qtail
* Modifies:	A[A] C[A] D[A1] D1
* Stack:	0
**********************************************************************

QSKIPD1		P=	10-1
		DSL	WP
		DSL	WP
		DSL	WP
		DSL	WP
		DSL	WP		D[A]=complevel=0
		P=	0
qskip1loop	A=DAT1	A
		D1=D1+	5
		LC(5)	=DOHSTR
		?A=C	A
		GOYES	qskip1hstr
		LC(2)	=DOLIST
		?A#C	A
		GOYES	qskip1semi
		D=D+1	A		complevel++
		GONC	qskip1loop
qskip1hstr	A=DAT1	A
		CD1EX
		C=C+A	A
		CD1EX
		?D#0	A
		GOYES	qskip1loop
		GONC	qskipend
qskip1semi	D=D-1	A
		?D#0	A
		GOYES	qskip1loop
		GONC	qskipend

**********************************************************************
* Desc:		Count the number of variables in a polynomial
* Entry:	D0 = ->poly
* Exit:		A[A] = B[A] = vars	D0 = ->polytail
* Alters:	A[A] B[A] C[A] D[A]
**********************************************************************
QVARS		B=0	A		vars = 0
		A=DAT0	A
		LC(5)	=DOLIST
		?A#C	A
		GOYES	cntvarok
		D=0	A		complevel = 0

cntvarlist	D0=D0+	5
		D=D+1	A		complevel++
		C=D	A
		?C<=B	A
		GOYES	cntvarloop
		B=C	A		vars = complevel (max so far)
cntvarloop	A=DAT0	A
		LC(5)	=DOLIST
		?A=C	A
		GOYES	cntvarlist
		D0=D0+	5		Skip DOHSTR/SEMI
		LC(5)	=SEMI
		?A=C	A
		GOYES	cntvarsemi
		C=DAT0	A		Zlen
		AD0EX
		A=A+C	A		Skip Z
		D0=A
		GONC	cntvarloop
cntvarsemi	D=D-1	A		complevel--
		?D#0	A
		GOYES	cntvarloop
cntvarok	A=B	A		vars
		RTN
**********************************************************************
* Desc:		Count the number of variables in two polynomials on stack
* Entry:	D1 = ->stack level 1		( P1 P2 )
* Exit:		A[A] = max[vars(P1),vars(P2)]
*		B[A] = vars(P1)
*		C[A] = vars(P2)
* Alters:	A[A] B[A] C[A] D[A]
* Stack:	2
**********************************************************************
QVARS2		A=DAT1	A
		D0=A			D0 = ->P2
		GOSUB	QVARS
		C=A	A
		RSTK=C			RSTK = vars2
		D1=D1+	5
		A=DAT1	A
		D1=D1-	5
		D0=A			D0 = ->P1
		GOSUB	QVARS		A[A] = vars1
		C=RSTK			C[A] = vars2
		?A>=C	A
		RTNYES			vars1 is greater
		A=C	A		vars2 is greater
		RTN

**********************************************************************
* Desc:		Fast equality test for objects
* Entry:	D0 = ->ob1
*		D1 = ->ob2
* Exit:		D0/D1 unchanged
*		CC: Equal
*		CS: Not equal
* Modifies:	A[W] C[W] B[A] D[A] ST[1]
* Stack:	3 (SKIPOB call)
**********************************************************************

EQUAL?

* First try to resolve the matter quickly by comparing addresses

		AD0EX
		D0=A
		CD1EX
		D1=C
		?A#C	A
		GOYES	+
		RTNCC			CC: Equal address

* Then by comparing first 7 nibbles (DOCHAR size)

+		A=DAT0	7
		C=A	W		Avoid WP test this way
		C=DAT1	7
		?A#C	W
		RTNYES			CS: First 10 nibbles different

* Then compare sizes

		AD0EX
		B=A	A		->ob1
		D0=A
		GOSBVL	=SKIPOB
		CD0EX
		C=C-B	A		size1
		CBEX	A		B[A]=size1
		CD1EX
		D0=C			->ob2
		D=C	A		->ob2
		GOSBVL	=SKIPOB
		AD0EX			->ob2 tail
		C=D	A		->ob2
		A=A-C	A		size2
		CD1EX			->ob2
		D0=C			->ob1
		?A#B	A
		RTNYES			CS: Different size

* Finally compare the equal length bodies

		CD0EX			Save start addresses
		RSTK=C			->ob1
		D0=C
		CD1EX
		RSTK=C			->ob2
		D1=C

		C=B	A		nibbles
		BSR	A		words
		P=C	0		nibbles&15

equalWloop	A=DAT0	W
		C=DAT1	W
		B=B-1	A		Compare words
		GOC	equalWP?
		D0=D0+	16
		D1=D1+	16
		?A=C	W
		GOYES	equalWloop
equalnot	P=	0		Restore regs and return CS
		C=RSTK
		D1=C
		C=RSTK
		D0=C
		RTNSC
equalWP?	P=P-1
		GOC	isequal		we had even words - test done
		?A#C	WP
		GOYES	equalnot
isequal		P=	0		restore regs and return CC
		C=RSTK
		D1=C
		C=RSTK
		D0=C
		RTNCC

**********************************************************************
* Desc:		Test if ob1 contains ob2
* Entry:	D0 = ->ob1
*		D1 = ->ob2
* Exit:		CC if ob1 contains ob2, else CS
*		For lists R0[A] contains match position or zero
* Alters:	A[W] C[W] B[A] D[A] D0 D1 R0[W] R1[W] ST[1]
* Stack:	4 (EQUAL? call)
* Notes:	Special care is taken to include ROM objects in
*		the comparisons.
**********************************************************************
CONTAINS?	A=DAT0	A
		LC(5)	=DOLIST		#02A74
		?A=C	A
		GOYES	listcontains

		LC(2)	=DOSYMB		#02AB8
		?A=C	A
		GOYES	symbcontains

* Compare regular objects

		GOTO	EQUAL?		CC if match

* Search list for a match

listcontains	A=0	A
		R0=A	A		position = 0
		D0=D0+	5
-		A=R0	A
		A=A+1	A		position++
		R0=A	A
		A=DAT0	A
		LC(5)	=SEMI
		?A=C	A
		GOYES	listnocontain	End of list - no match then
		AD0EX
		R1=A	A		R1[A] = ->ptr
		A=DAT0	A
		LC(5)	=PRLG
		?A#C	A
		GOYES	+		Indirect compare
		A=R1	A
		D0=A
+		GOSUB	EQUAL?
		RTNNC			If match then return CC
		A=R1	A		Restore ->ptr
		D0=A
		GOSBVL	=SKIPOB		Else skip object
		GOTO	-		and loop
listnocontain	A=0	A
		R0=A	A		Return 0 as match position
		RTNSC			CS: No match


* Search symbolic for a match

symbcontains	A=0	A
		R0=A	A		R0[A] = depth
-		A=DAT0	A
		AD0EX
		R1=A	A		R1[A] = ->ptr
		A=DAT0	A
		LC(5)	=PRLG
		?A#C	A		Indirect compare
		GOYES	+
		A=R1	A
		D0=A
+		GOSUB	EQUAL?
		RTNNC			If match then return CC
		A=R1	A		Restore ->ptr
		D0=A
		A=DAT0	A
		LC(5)	=DOSYMB
		?A#C	A
		GOYES	symbconsymb
		D0=D0+	5
		A=R0	A
		A=A+1	A		depth++
		R0=A	A
		GONC	+		In case of null symbolic
symbconsymb	GOSBVL	=SKIPOB		Skip object
+		A=DAT0	A
		LC(5)	=SEMI
		?A#C	A
		GOYES	-		And continue if not SEMI
		D0=D0+	5
		A=R0	A
		A=A-1	A		depth--
		R0=A	A
		?A#0	A
		GOYES	-
		RTNSC			Traversed all - no match = CS


**********************************************************************
* Desc:		General short Z push & exit entries.
* Entry:	A[W] mantissa
*		C[S] sign
**********************************************************************

* Entry to drop 3 before pushing

PushN3Loop	C=0	S		Positive
PushZ3Loop	GOSUB	PushZ		Push
		A=DAT1	A		->Z
		D1=D1+	5		And do 4UNROLL3DROP
		D=D+1	A
		GONC	+

* Entry to drop 2 before pushing

PushN2Loop	C=0	S		Positive
PushZ2Loop	GOSUB	PushZ		Push
		A=DAT1	A		->Z
+		D1=D1+	5		And do UNROT2DROP
		D=D+1	A
		GONC	+

* Entry to drop 1 before pushing

PushN1Loop	C=0	S		Positive
PushZ1Loop	GOSUB	PushZ		Push
		A=DAT1	A		->Z
+		D1=D1+	5
		D=D+1	A
		DAT1=A	A
		GONC	+

* Entry to simply push

PushNLoop	C=0	S		Positive
PushZLoop	GOSUB	PushZ		Push
+		GOVLNG	=Loop	

**********************************************************************
* Desc:		Push Z to data stack safely
* Entry:	A[W] mantissa
*		C[S] sign
* Uses:		IRAMBUFF to save during GC
*		A,B,C,D,D0,D1 sGARB
* Note:		This can safely be called from a subroutine, but not deeper.
**********************************************************************

PushZ0		LC(5)	11
		GOSBVL	=GETTEMP
		LCZ0
		DAT0=C	11
		AD0EX
		GOVLNG	=GPPushA

PushN		C=0	S		Special entry to push positive
PushZ		?A=0	W
		GOYES	PushZ0
		P=	0
-		P=P-1
		?A=0	P
		GOYES	-
		D0=(5)	(=IRAM@)-4
		C=DAT0	A
		D0=C
		D0=(4)	#100		IRAMBUFF on SX and GX
		DAT0=C	S		sign
		D0=D0+	1
		DAT0=A	W		mantissa
		C=0	A
		CPEX	0		nibbles-1
		C=C+CON	A,12		Z size
		GOSBVL	=GETTEMP	C[A]=size  D0=->ob	D1=->obend
		B=C	A		Z size
		CD0EX
		D0=(5)	(=IRAM@)-4
		A=DAT0	A
		D0=A
		D0=(4)	#100		IRAMBUFF on SX and GX
		C=DAT0	S		sign
		D0=D0+	1
		A=DAT0	W		mantissa
		D0=C			->ob
		D1=D1-	1
		DAT1=C	S		Write sign
		LC(5)	=DOHSTR
		DAT0=C	A		Write prolog
		D0=D0+	5
		C=B	A		Z size
		C=C-CON	A,5
		DAT0=C	A		Write length
		D0=D0+	5
		C=C-CON	A,7
		P=C	0
		DAT0=A	WP		Write mantissa
		P=	0
		D0=D0-	10
		AD0EX
		GOVLNG	=GPPushA

PushZ1		LC(5)	12
		GOSBVL	=GETTEMP
		LCZ1
		DAT0=C	12
		AD0EX
		GOVLNG	=GPPushA

**********************************************************************
* Desc:		Univariate test on polynomial
* Entry:	D0 = ->Q
* Exit:		CS: Is univariate
*		CC: Not univariate
* Modifies:	A[A] C[A] D0 sUNIVAR
* Stack:	0
**********************************************************************

sUNIVAR		EQU 5		Flag to indicate plain univariate test

UNIVAR?		ST=1	sUNIVAR		Flag plain univariate test
		GOTO	unitest

**********************************************************************
* Desc:		Univariate and short coefficient test on polynomial
* Entry:	D0 = ->Q
* Exit:		CS: Is univariate and short
*		CC: Not
* Modifies:	A[A] C[A] D0 sUNIVAR
* Stack:	0
**********************************************************************

SUNIVAR?	ST=0	sUNIVAR		Flag coefficient bound test

unitest		A=DAT0	A
		LC(5)	=DOLIST
		?A#C	A
		RTNYES			CS - non-list is univariate
		D0=D0+	5
-		A=DAT0	A
		LC(5)	=SEMI		If we reach SEMI
		?A=C	A
		RTNYES			then P is sunivariate
		LC(4)	=DOHSTR		If there is an embedded object
		?A#C	A
		GOYES	notunivar	then P is not sunivariate
		D0=D0+	5
		A=DAT0	A		Zlen
		CD0EX
		C=C+A	A
		CD0EX
		?ST=1	sUNIVAR		Ignore coefficient test?
		GOYES	-		Yes
		LC(5)	5+15
		?A<=C	A
		GOYES	-
notunivar	RTNCC			CC - not univariate

**********************************************************************
* Desc:		Short coefficient test on polynomial
* Entry:	D0 = ->Q
* Exit:		CS: Is short
*		CC: Not
* Modifies:	A[A] C[A] D[A] D0
* Stack:	0
**********************************************************************
QSHORT?		D=0	A		D[A] = depth = 0
-		A=DAT0	A
		D0=D0+	5
		LC(5)	=DOHSTR
		?A=C	A
		GOYES	qshortz
		LC(2)	=DOLIST
		?A#C	A
		GOYES	qshorts
		D=D+1	A		depth++
		GONC	-
qshorts		D=D-1	A		depth--
		?D#0	A
		GOYES	-
		RTNSC			CS: Is short
qshortz		A=DAT0	A		A[A] = Zlen
		CD0EX
		C=C+A	A		Skip Z
		D0=C
		LC(5)	5+15+1		Mantissa+sign must fit into word
		?A<=C	A
		GOYES	-		This does - continue
		RTNCC			CC: Not short

**********************************************************************
* Name:		QSHORT2?
* Desc:		Test if stack contains two short polynomials
* Entry:	D1 = ->stack level 1		( P1 P2 )
* Exit:		CS: Both short
* Alters:	A[A] C[A] D[A] D0
**********************************************************************
QSHORT2?	A=DAT1	A		->P2
		D0=A
		GOSUB	QSHORT?		CS if P2 is short
		RTNNC			CC: Not
		D1=D1+	5
		A=DAT1	A		->P1
		D1=D1-	5
		D0=A
		GOTO	QSHORT?		CS if P1 is short

**********************************************************************
* Desc:		Shrink allocated memory to contain only Q
* Entry:	R0[A] = ->Q
*		C[A] = ->Qtail
* Exit:		GETPTR and A[A]=R0[A] = ->Q
**********************************************************************
ShrinkQ		D0=C			D0 = ->Qtail
		A=R0	A		A[A] = ->Q
		C=C-A	A		C[A] = Qsize
		C=C+CON	A,6
		DAT0=C	A		Fix link
		D0=D0+	5		Skip link
		GOSBVL	=recover	Shrink allocated memory
		A=R0	A		A[A] = ->Q
		GOVLNG	=GETPTR
**********************************************************************
* Desc:		Trash topmost object in tempob area
* Entry:	RPLRAM
* Exit:		RPLRAM'
**********************************************************************
?TrashTop	?ST=0	sTRASH		Optional entry which obeys
		RTNYES			the global trash flag
TrashTop	D0=(5)	=aTEMPTOP
		A=DAT0	A
		D0=A
		A=DAT0	A		A[A] = ->top of tempob area
		A=A-CON	A,5		A[A] = ->link1
		D0=A
		C=DAT0	A		C[A] = link1
		A=A-C	A		A[A] = ->link2
		D0=A
		D0=D0+	5		D0 = ->top of tempob (desired)
		GOVLNG	=recover	Trash topmost object

**********************************************************************
* Desc:		Generate abort error if ON-key has been pressed
* Entry:	RPLCPU
* Alters:	A[A]
**********************************************************************
AttnAbort?	CD0EX
		D0=(5)	=aATTNFLG
		A=DAT0	A		A[A] = ->ATTNFLG
		D0=A
		A=DAT0	A		A[A] = ATTNFLG
		CD0EX
		?A=0	A
		RTNYES
		A=0	A		0 = ABORT error
		CD0EX
		DAT0=A	A		ATTNFLG = 0
		CD0EX
		GOVLNG	=Errjmp
**********************************************************************
RPL



