**********************************************************************
* Name:		FastLib
* Stack:	( --> )
* Desc:		Provides library speed-up by poking the internal
*		library table.
* Notes:
*	Individual romps are executed by 1st searching for the corresponding
*	lid from the internal ROMPTAB, then fetching the link table for the
*	library, and finally fetching the program body from the address
*	referred to by the link table.
*	ROMPTAB is sorted by LID, from 0 up to #7FF. To speed things up
*	this program removes the sort by overwriting the useless slot
*	for library 0 with the true slot in the table. The slot info
*	is then restored at the end of program execution, the restore
*	program is swapped with the calling stream to achieve this.
*
*	Unfortunately the romp execution assumes the ROMPTAB is sorted and
*	will stop searching for the proper LID if a bigger LID is encountered.
*
*	THUS ANY LIBRARY WHOSE LID IS SMALLER THAN FROMID CANNOT BE FOUND
*	VIA THE NORMAL ROMPTR EXECUTION.
*
*	The Profiler works only for libraries that don't modify the run
*	stream. So that this program alone wouldn't make the Profiler error
*	the code pushing the errortrap to the stream will check if the
*	charasteristic code for the Profiler exists in the calling stream, if
*	so the errortrap is pushed below the two topmost streams. To keep
*	things simple I chose a following code object in the stream to mean
*	the Profiler is running. If your original code has this too right
*	where you want to insert the FastLib call then just insert a NOP
*	command in between.
*
*	Note that the error trap is completely in machine language so
*	that not even an out-of-memory error can prevent the trap from
*	working. Also note that a Warmstart rebuilds the internal table
*	so that warmstart is no problem either.
*
*	FastLib also checks whether the library it is in is covered, and
*	if so it will uncover the proper bank temporarily so that any
*	future ROMPTR executions will see the library. This enables
*	using normally unsafe memory allocation calls. However the trick
*	also means that calling any subroutine at all which might access
*	any memory bank at all is strictly forbidden.
*
*	To enable external libraries to call ALG48 subroutines any
*	library residing in the same bank will also be 'uncovered' by
*	changing the access field to zero value.
*
*	The errortrap pushed in the return stack will also restore the
*	card configuration to its proper state.
*
*	The command should be used by inserting the FastLib call in a
*	suitable place in the topmost level. So you would change the commands
*	like this: 
*
*		::
*			CK1&Dispatch
*			FIVE	:: FastLib <original dispatchee> ;
*			FOUR	:: FastLib <original dispatchee> ;
*			ONE	:: FastLib NOP <code> ;
*			...		   ^
*		;			   +------ See the profiler exception
*
*	ROMPTAB structure:	GX Slots:		SX Slots:
*
*	CON(3)	N		CON(3)	LID		CON(3)	LID
*	Slot1			CON(5)	LibAddress	CON(5)	LibAddress
*	Slot2			CON(5)	Access
*	...			CON(3)	Unused
*	SlotN
*
*	Error conditions:
*
*		Library is covered but card 1 is merged.
*
*	Warning:
*
*		Any object referenced in card 1 could cause a crash
*		if used while card 1 is unconnfigured.
*		
**********************************************************************

ASSEMBLE

* Status flag to mark the restore code must excute ERRJMP

sFERROR		EQU 5

* Library configuration table addresses

=ROMPTAB	EQU #707D9 * V
=G_ROMPTAB	EQU #809A3 * V

* Configuration table to check for marged card 1

=G_CONFTAB	EQU #8052B * V

* Selftest is invoked by ON-D, and thus the RAM variables used by
* it are suitable as long-period save locations. FastLib uses it to save
* old library 0 configuration info and the access pointer address for
* the current library

=TESTMSG	EQU #7008C * V	TESTMSG is the text buffer used by ON-D for
=G_TESTMSG	EQU #80092 * V	building display lines (22 characters)

=FINDLAM	EQU #07DA2 * U  Find lambda variable given name

* Bank uncover disables ON-C and similar keys by clearing INITEN
* FastLib enables it (and the other combinations) again as tests prove
* they work just fine.

=G_INITEN	EQU #8065A * V

* Provide the library number in a convenient way so we don't have to worry
* about forgetting to change any romid equates.
* Note that the prologue will exist even if FastLib is in TEMPOB!

fastlibprlg	CON(6)	=~FastLib
RPL
NULLNAME FastLib
::

* First see if FastLib is disabled by having SDB running

  CODE
		GOSBVL	=SAVPTR
		D0=(5)	=IRAM@		Create LAM name form into TESTMSG area
		C=DAT0	S
		D0=(5)	=G_TESTMSG
		C=C+C	S
		GOC	sdbgx?
		D0=(5)	=TESTMSG
sdbgx?		LC(N)	2+2*4
		CON(2)	4		4 chars
		NIBASC	\~dbg\		lam name used by SDB
		DAT0=C	2+2*4
		AD0EX
		R0=A	A		->name form
		GOSBVL	=FINDLAM	CC: D0 = ->lam binding
		GOVLNG	=GPPushT/FLp	CS: No match
  ENDCODE
  NOT?SEMI				( Return if SDB is running )

  ERRSET
  ::

     CODE
		GOSBVL	=SAVPTR

* Fetch library number

		A=PC
		LC(5)	(fastlibprlg)-(*)
		A=A+C	A
		D0=A
		C=DAT0	X		LID

* Dispatch for separate code for SX and GX

		D0=(5)	=IRAM@
		C=DAT0	S
		C=C+C	S
		GOC	fastlibgx

******************************************************
* Make SX library fast, no worries about being covered
******************************************************
fastlibsx	D0=(5)	=ROMPTAB
		A=DAT0	A
		D0=D0+	3
		B=A	A		B[X] = N libraries
sxfindlp	B=B-1	X
		GOC	sxfindfail	IMPOSSIBLE!?
		A=DAT0	A
		D0=D0+	8
		?A#C	X
		GOYES	sxfindlp
		D0=D0-	8		->LID slot

		A=DAT0	8		LIB info
		D0=(5)	(=ROMPTAB)+3
		C=DAT0	8		LIB0 info
		DAT0=A	8		Replace with LIB info
		D0=(5)	=TESTMSG
		DAT0=C	8		Save LIB0 info

		GOTO	pushfastlib	Reorder RPL return stack

* This must not reach the error trap below, thus it must do RDROP first

sxfindfail	GOSBVL	=GETPTR
		B=B-CON	A,5
		D=D+1	A
		LC(5)	#12D		"Missing Library"
		GOVLNG	=ErrjmpC

*****************************************************
* Make GX library fast, and uncover bank if necessary
*****************************************************

uncoverbank	P=	0		"Uncover"
		PC=C

fastlibgx	D0=(5)	=G_ROMPTAB
		A=DAT0	A
		D0=D0+	3
		B=A	A		B[X] = N libraries
gxfindlp	B=B-1	X
		GOC	sxfindfail	IMPOSSIBLE!?
		A=DAT0	A
		D0=D0+	16
		?A#C	X
		GOYES	gxfindlp

* If lib is covered and card 1 is merged, error

		D0=D0-	3+5
		C=DAT0	A
		D0=D0-	3+5		->LID slot
		?C=0	A
		GOYES	+		Fine - not covered

		D1=(5)	(=G_CONFTAB)+4
		C=DAT1	B
		?CBIT=1	1		SystemRam? ([r w s 0])
		GOYES	sxfindfail	"Missing Library" - well..

* Card 1 not merged - continue

+		A=DAT0	W		LIB info
		D0=(5)	(=G_ROMPTAB)+3
		C=DAT0	W		LIB0 info
		DAT0=A	W		Replace with LIB info
		D0=(5)	=G_TESTMSG
		DAT0=C	W		Save LIB0 info
		D0=D0+	16
		DAT0=A	W		Save LIB info (access pointer!)

* Now if the library is covered, uncover it

		D0=D0+	3+5
		C=DAT0	A		access
		?C=0	A
		GOYES	pushfastlib	Already visible - reorder return stack

		GOSUB	uncoverlibs	Uncover libs with same access
		GOSUB	uncoverbank	Uncover the bank
*		D0=(5)	(=G_ROMPTAB)+3+3+5
*		A=0	A
*		DAT0=A	A		'uncover'
		D0=(5)	=G_INITEN	And enable ON-C etc again
		LC(1)	#F
		DAT0=C	P

*****************************************************************
* Now reorder RPL return stack so that the errortrap is on bottom
*****************************************************************
pushfastlib	GOSBVL	=GETPTR
		A=B	A
		D0=A
		D0=D0-	10
		A=DAT0	A		->calling stream
		D1=A
		A=DAT1	A		See if Profiler is fiddling the stream
		LC(5)	=DOCODE
		?A=C	A
		GOYES	fastProfile
* No profiler, just do RSWAP
		A=DAT0	A
		D0=D0+	5
		C=DAT0	A
		DAT0=A	A
		D0=D0-	5
		DAT0=C	A
		GONC	fastexit
* Found profiler, do RUNROT
fastProfile	D0=D0+	5
		A=DAT0	A
		D0=D0-	10
		C=DAT0	10
		DAT0=A	A
		D0=D0+	5
		DAT0=C	10
fastexit	GOVLNG	=GETPTRLOOP
     ENDCODE
  ;

  ERRTRAP
* The trap
  CODE
fastlibtrap	ST=1	sFERROR
		GOTO	fastrestore
  ENDCODE

* Make sure output is not covered

  ERRSET				( To protect TOTEMPOB )
  CODE
		A=DAT1	A		[Empty stack would give zero]
		?A>B	A		If ob above return stack then must
		GOYES	+		be in the library itself
		GOVLNG	=Loop
+		LC(5)	=TOTEMPOB	So do an unconditional TOTEMPOB
		A=C	A
		PC=(A)
  ENDCODE
  ERRTRAP
  CODE
		GOTO	fastlibtrap	TOTEMPOB failed - error
  ENDCODE

* Normal restore

  CODE
		ST=0	sFERROR
fastrestore	GOSBVL	=SAVPTR

* Dispatch for separate code for SX and GX

		D0=(5)	=IRAM@
		C=DAT0	S
		C=C+C	S
		GOC	restoregx

***************************************
* Restore original ROMPTAB state for SX
***************************************
restoresx	D0=(5)	=TESTMSG
		A=DAT0	8			LIB0 info
		D0=(5)	(=ROMPTAB)+3
		DAT0=A	8			Restore it
		GOTO	restoreexit

************************************************
* Restore original ROMPTAB and BANK state for GX
************************************************
coverbank	P=	1			"Cover"
		PC=C

restoregx	D0=(5)	(=G_TESTMSG)+16+3+5
		C=DAT0	A			access
		?C=0	A
		GOYES	skipcover
		GOSUB	coverlibs		Cover the libraries
		GOSUB	coverbank		Cover the bank
skipcover	D0=(5)	=G_TESTMSG		Restore LIB0 info after
		A=DAT0	W			coverlibs to avoid changing
		D0=(5)	(=G_ROMPTAB)+3		LIB0 access address!!
		DAT0=A	W

***************************************************
* Now error or continue based on entry into restore
***************************************************
restoreexit	?ST=1	sFERROR
		GOYES	ferror
		GOVLNG	=GETPTRLOOP
ferror		LC(5)	=ERRJMP
		GOVLNG	=GETPTREVALC


**************************************************
* Clear access fields of libraries in a port
* C[A] = access (not modified)
*	CON(3)	lid		CON(3)	lid
*	CON(5)	addr	-->	CON(5)	addr
*	CON(5)	access		CON(5)	0
*	CON(3)	extra		CON(3)	nonzero
**************************************************
uncoverlibs	D0=(5)	=G_ROMPTAB
		A=DAT0	A
		B=A	A		B[X]=libs
-		D0=D0+	3+3+5
		B=B-1	X
		RTNC			Done all
		A=DAT0	A
		?A#C	A		Different port?
		GOYES	+
		A=0	A		Uncover lib
		DAT0=A	A
		D0=D0+	5
		A=A+1	A
		DAT0=A	X
		GONC	-
+		D0=D0+	5		Don't uncover
		A=0	A
		DAT0=A	X
		GONC	-

**************************************************
* Restore access fields of libraries in a port
* C[A] = access (not modified)
*	CON(3)	lid		CON(3)	lid
*	CON(5)	addr	-->	CON(5)	addr
*	CON(5)	0		CON(5)	access
*	CON(3)	nonzero		CON(3)	nonzero
**************************************************
coverlibs	D0=(5)	=G_ROMPTAB
		A=DAT0	A
		B=A	A		B[X]=libs
-		D0=D0+	3+3+5+5		->extra
		B=B-1	X
		RTNC			done all
		A=DAT0	A
		D0=D0-	5		->access
		?A=0	X
		GOYES	+		ignore
		DAT0=C	A		fix back
+		D0=D0+	5
		GONC	-
  ENDCODE
;
**********************************************************************



